from __future__ import annotations

from typing import Protocol

from playwright.sync_api import Locator, Page
from playwright.sync_api import expect as playwright_expect

from .._types import PatternOrStr, StyleValue, Timeout
from ..expect._internal import expect_style_to_have_value as _expect_style_to_have_value
from ._base import OutputBaseP, UiBaseP, UiWithContainer, WidthLocM


class _CardBodyP(UiBaseP, Protocol):
    """
    Represents the body of a card control.
    """

    loc_body: Locator
    """
    Playwright `Locator` for the body element of the card control.
    """


class _CardBodyM:
    """Represents a card body element with additional methods for testing."""

    def expect_body(
        self: _CardBodyP,
        value: PatternOrStr | list[PatternOrStr],
        *,
        timeout: Timeout = None,
    ) -> None:
        """Expect the card body element to have the specified text.

        Parameters
        ----------
        value
            The expected text or a list of expected texts.
        timeout
            The maximum time to wait for the text to appear. Defaults to `None`.
        """
        playwright_expect(self.loc).to_have_text(
            value,
            timeout=timeout,
        )


class _CardFooterLayoutP(UiBaseP, Protocol):
    """
    Represents the layout of the footer in a card.
    """

    loc_footer: Locator
    """
    Playwright `Locator` for the footer element.
    """


class _CardFooterM:
    """
    Represents the footer section of a card.
    """

    def expect_footer(
        self: _CardFooterLayoutP,
        value: PatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Asserts that the footer section of the card has the expected text.

        Parameters
        ----------
        value
            The expected text in the footer section.
        timeout
            The maximum time to wait for the footer text to appear. Defaults to `None`.
        """
        playwright_expect(self.loc_footer).to_have_text(
            value,
            timeout=timeout,
        )


class _CardValueBoxFullScreenLayoutP(OutputBaseP, Protocol):
    """
    Represents a card / Value Box full-screen layout for the Playwright controls.
    """

    loc_title: Locator
    """
    Playwright `Locator` for the title element.
    """
    _loc_fullscreen: Locator
    """
    Playwright `Locator` for the full-screen element.
    """
    _loc_close_button: Locator
    """
    Playwright `Locator` for the close button element.
    """


class _CardValueBoxFullScreenM:
    """
    Represents a class for managing full screen functionality of a Card or Value Box.
    """

    def set_full_screen(
        self: _CardValueBoxFullScreenLayoutP, open: bool, *, timeout: Timeout = None
    ) -> None:
        """
        Sets the element to full screen mode or exits full screen mode.

        Parameters
        ----------
        open
            `True` to open the element in full screen mode, `False` to exit full screen mode.
        timeout
            The maximum time to wait for the operation to complete. Defaults to `None`.
        """
        if open:
            self.loc_title.hover(timeout=timeout)
            self._loc_fullscreen.wait_for(state="visible", timeout=timeout)
            self._loc_fullscreen.click(timeout=timeout)
        else:
            self._loc_close_button.click(timeout=timeout)

    def expect_full_screen(
        self: _CardValueBoxFullScreenLayoutP, value: bool, *, timeout: Timeout = None
    ) -> None:
        """
        Verifies if the full screen mode is currently open.

        Parameters
        ----------
        value
            `True` if the item is to be in full screen mode, `False` otherwise.
        timeout
            The maximum time to wait for the verification. Defaults to `None`.
        """
        playwright_expect(self._loc_close_button).to_have_count(
            int(value), timeout=timeout
        )

    def expect_full_screen_available(
        self: _CardValueBoxFullScreenLayoutP,
        value: bool,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects whether full screen mode is available for the element.

        Parameters
        ----------
        value
            `True` if the element is expected to be available for full screen mode, False otherwise.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        playwright_expect(self._loc_fullscreen).to_have_count(
            int(value), timeout=timeout
        )


class ValueBox(
    WidthLocM,
    _CardValueBoxFullScreenM,
    UiWithContainer,
):
    """
    Controller for :func:`shiny.ui.value_box`.
    """

    loc: Locator
    """
    Playwright `Locator` for the value box's value.
    """
    loc_showcase: Locator
    """
    Playwright `Locator` for the value box showcase.
    """
    loc_title: Locator
    """
    Playwright `Locator` for the value box title.
    """
    loc_body: Locator
    """
    Playwright `Locator` for the value box body.
    """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `ValueBox` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the value box.

        """
        super().__init__(
            page,
            id=id,
            loc_container=f"div#{id}.bslib-value-box",
            loc="> div.card-body > .value-box-grid, > div.card-body",
        )
        value_box_grid = self.loc
        self.loc_showcase = value_box_grid.locator("> .value-box-showcase")
        self.loc_title = value_box_grid.locator("> .value-box-area > .value-box-title")
        self.loc = value_box_grid.locator("> .value-box-area > .value-box-value")
        self.loc_body = value_box_grid.locator(
            "> .value-box-area > :not(.value-box-title, .value-box-value)"
        )
        self._loc_fullscreen = self.loc_container.locator(
            "> bslib-tooltip > .bslib-full-screen-enter"
        )

        # an easier approach is using `#bslib-full-screen-overlay:has(+ div#{id}.card) > a`
        # but playwright doesn't allow that
        self._loc_close_button = (
            self.page.locator(f"#bslib-full-screen-overlay + div#{id}.bslib-value-box")
            .locator("..")
            .locator("#bslib-full-screen-overlay > a")
        )

    def expect_height(self, value: StyleValue, *, timeout: Timeout = None) -> None:
        """
        Expects the value box to have a specific height.

        Parameters
        ----------
        value
            The expected height value.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        _expect_style_to_have_value(
            self.loc_container, "max-height", value, timeout=timeout
        )

    def expect_title(
        self,
        value: PatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the value box title to have a specific text.

        Parameters
        ----------
        value
            The expected text pattern or string.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.

        """
        playwright_expect(self.loc_title).to_have_text(
            value,
            timeout=timeout,
        )

    def expect_value(
        self,
        value: PatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the value box value to have a specific text.

        Parameters
        ----------
        value
            The expected text pattern or string.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        playwright_expect(self.loc).to_have_text(
            value,
            timeout=timeout,
        )

    def expect_body(
        self,
        value: PatternOrStr | list[PatternOrStr],
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the value box body to have specific text.

        Parameters
        ----------
        value
            The expected text pattern or list of patterns/strings.

            Note: If testing against multiple elements, text should be an array.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        playwright_expect(self.loc_body).to_have_text(
            value,
            timeout=timeout,
        )


class Card(
    WidthLocM,
    _CardFooterM,
    _CardBodyM,
    _CardValueBoxFullScreenM,
    UiWithContainer,
):
    """
    Controller for :func:`shiny.ui.card`.
    """

    loc_container: Locator
    """
    Playwright `Locator` for the card container.
    """
    loc: Locator
    """
    Playwright `Locator` for the card's value.
    """
    loc_title: Locator
    """
    Playwright `Locator` for the card title.
    """
    loc_footer: Locator
    """
    Playwright `Locator` for the card footer.
    """
    loc_body: Locator
    """
    Playwright `Locator` for the card body.
    """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `Card` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the card.
        """
        super().__init__(
            page,
            id=id,
            loc_container=f"div#{id}.card",
            loc="> div.card-body",
        )
        self.loc_title = self.loc_container.locator("> div.card-header")
        self.loc_footer = self.loc_container.locator("> div.card-footer")
        self._loc_fullscreen = self.loc_container.locator(
            "> bslib-tooltip > .bslib-full-screen-enter"
        )
        # an easier approach is using `#bslib-full-screen-overlay:has(+ div#{id}.card) > a`
        # but playwright doesn't allow that
        self._loc_close_button = (
            self.page.locator(f"#bslib-full-screen-overlay + div#{id}")
            .locator("..")
            .locator("#bslib-full-screen-overlay > a")
        )
        self.loc_body = self.loc

    def expect_header(
        self,
        value: PatternOrStr | None,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the card header to have a specific text.

        Parameters
        ----------
        value
            The expected text pattern or string.

            Note: `None` if the header is expected to not exist.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        if value is None:
            playwright_expect(self.loc_title).to_have_count(0, timeout=timeout)
        else:
            playwright_expect(self.loc_title).to_have_text(value, timeout=timeout)

    # def expect_body(
    #     self,
    #     text: PatternOrStr,
    #     index: int = 0,
    #     *,
    #     timeout: Timeout = None,
    # ) -> None:
    #     """Note: Function requires an index since multiple bodies can exist in loc"""
    #     playwright_expect(self.loc.nth(index).locator("> :first-child")).to_have_text(
    #         text,
    #         timeout=timeout,
    #     )

    def expect_footer(
        self,
        value: PatternOrStr | None,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the card footer to have a specific text.

        Parameters
        ----------
        value
            The expected text pattern or string
            Note: None if the footer is expected to not exist.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        if value is None:
            playwright_expect(self.loc_footer).to_have_count(0, timeout=timeout)
        else:
            playwright_expect(self.loc_footer).to_have_text(value, timeout=timeout)

    def expect_max_height(self, value: StyleValue, *, timeout: Timeout = None) -> None:
        """
        Expects the card to have a specific maximum height.

        Parameters
        ----------
        value
            The expected maximum height value.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        _expect_style_to_have_value(
            self.loc_container, "max-height", value, timeout=timeout
        )

    def expect_min_height(self, value: StyleValue, *, timeout: Timeout = None) -> None:
        """
        Expects the card to have a specific minimum height.

        Parameters
        ----------
        value
            The expected minimum height value.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        _expect_style_to_have_value(
            self.loc_container, "min-height", value, timeout=timeout
        )

    def expect_height(self, value: StyleValue, *, timeout: Timeout = None) -> None:
        """
        Expects the card to have a specific height.

        Parameters
        ----------
        value
            The expected height value.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        _expect_style_to_have_value(
            self.loc_container, "height", value, timeout=timeout
        )



from __future__ import annotations

__all__ = ("input_dark_mode", "update_dark_mode")

from typing import Literal, Optional

from htmltools import Tag, TagAttrValue, css

from .._docstring import add_example, no_example
from .._namespaces import resolve_id
from ..session import Session, require_active_session
from ._web_component import web_component

BootstrapColorMode = Literal["light", "dark"]


@add_example()
def input_dark_mode(
    *,
    id: Optional[str] = None,
    mode: Optional[BootstrapColorMode] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Creates a dark mode switch input that toggles the app between dark and light modes.

    Parameters
    ----------
    id
        An optional ID for the dark mode switch. When included, the current color mode
        is reported in the value of the input with this ID.
    mode
        The initial mode of the dark mode switch. By default or when set to `None`, the
        user's system settings for the preferred color scheme will be used. Otherwise,
        set to `"light"` or `"dark"` to force the initial mode.
    **kwargs
        Additional attributes to be added to the dark mode switch, such as `class_` or
        `style`.

    Returns
    -------
    :
        A dark mode toggle switch UI element.

    References
    ----------
    * <https://getbootstrap.com/docs/5.3/customize/color-modes>
    """

    if mode is not None:
        mode = validate_dark_mode_option(mode)

    if id is not None:
        id = resolve_id(id)

    return web_component(
        "bslib-input-dark-mode",
        {
            "style": css(
                **{
                    "--text-1": "var(--bs-emphasis-color)",
                    "--text-2": "var(--bs-tertiary-color)",
                    # TODO: Fix the vertical correction to work better with Bootstrap
                    "--vertical-correction": " ",
                },
            )
        },
        id=id,
        attribute="data-bs-theme",
        mode=mode,
        **kwargs,
    )


def validate_dark_mode_option(mode: BootstrapColorMode) -> BootstrapColorMode:
    if mode not in ("light", "dark"):
        raise ValueError("`mode` must be either 'light' or 'dark'.")
    return mode


@no_example()
def update_dark_mode(
    mode: BootstrapColorMode, *, session: Optional[Session] = None
) -> None:
    session = require_active_session(session)

    mode = validate_dark_mode_option(mode)

    msg: dict[str, object] = {
        "method": "toggle",
        "value": mode,
    }
    session._send_message_sync({"custom": {"bslib.toggle-dark-mode": msg}})



import random
import socketserver
from typing import List, Set

import pytest

from shiny._utils import AsyncCallbacks, Callbacks, private_seed, random_port
from shiny.ui._utils import extract_js_keys, js_eval


def test_randomness():
    current_state = random.getstate()
    try:
        # Make sure the public stream of randomness is independent of the private stream
        # https://github.com/posit-dev/py-shiny/issues/140
        pub = random.randint(0, 100000000)
        with private_seed():
            priv = random.randint(0, 100000000)
        pub2 = random.randint(0, 100000000)
        with private_seed():
            priv2 = random.randint(0, 100000000)
        assert pub != priv and priv != pub2 and pub2 != priv2

        # By setting the same seed, we should get the same randomness
        random.seed(0)
        public = [random.randint(0, 100000000) for _ in range(3)]
        with private_seed():
            random.seed(0)
            private = [random.randint(0, 100000000) for _ in range(3)]
        assert public == private

        # Interleaved calls to private and public should give the same randomness
        public: List[float] = []
        private: List[float] = []
        random.seed(0)
        with private_seed():
            random.seed(0)
        public.append(random.randint(0, 100000000))
        with private_seed():
            private.append(random.randint(0, 100000000))
        with private_seed():
            private.append(random.randint(0, 100000000))
        public.append(random.randint(0, 100000000))
        public.append(random.randint(0, 100000000))
        with private_seed():
            private.append(random.randint(0, 100000000))
        assert public == private

    finally:
        random.setstate(current_state)


def test_callbacks():
    class MockCallback:
        def __init__(self):
            self.exec_count = 0

        def __call__(self):
            self.exec_count += 1

    callbacks = Callbacks()

    cb1 = MockCallback()
    cb2 = MockCallback()
    cb3 = MockCallback()
    cb4 = MockCallback()

    def mutate_registrations():
        # Unregister cb2
        cb2_reg_handle()
        # Calling registrations multiple times should just no-op
        cb2_reg_handle()

        callbacks.register(cb4)

    _ = callbacks.register(cb1)
    mut_reg_handle = callbacks.register(mutate_registrations)
    cb2_reg_handle = callbacks.register(cb2)
    _ = callbacks.register(cb3, once=True)

    callbacks.invoke()
    assert cb1.exec_count == 1  # Regular registration was called
    assert cb2.exec_count == 1  # Unregistered by an earlier handler, but still called
    assert cb3.exec_count == 1  # once=True registration was called
    assert cb4.exec_count == 0  # Registered during invoke(), not called

    mut_reg_handle()

    callbacks.invoke()
    assert cb1.exec_count == 2  # Regular registration was called again
    assert cb2.exec_count == 1  # Unregistered by previous invoke, not called again
    assert cb3.exec_count == 1  # once=True, so not called again
    assert cb4.exec_count == 1  # Registered during previous invoke(), was called


@pytest.mark.asyncio
async def test_async_callbacks():
    class AsyncMockCallback:
        def __init__(self):
            self.exec_count = 0

        async def __call__(self):
            self.exec_count += 1

    callbacks = AsyncCallbacks()

    cb1 = AsyncMockCallback()
    cb2 = AsyncMockCallback()
    cb3 = AsyncMockCallback()
    cb4 = AsyncMockCallback()

    async def mutate_registrations():
        # Unregister cb2
        cb2_reg_handle()
        # Calling registrations multiple times should just no-op
        cb2_reg_handle()

        callbacks.register(cb4)

    _ = callbacks.register(cb1)
    mut_reg_handle = callbacks.register(mutate_registrations)
    cb2_reg_handle = callbacks.register(cb2)
    _ = callbacks.register(cb3, once=True)

    await callbacks.invoke()
    assert cb1.exec_count == 1  # Regular registration was called
    assert cb2.exec_count == 1  # Unregistered by an earlier handler, but still called
    assert cb3.exec_count == 1  # once=True registration was called
    assert cb4.exec_count == 0  # Registered during invoke(), not called

    mut_reg_handle()

    await callbacks.invoke()
    assert cb1.exec_count == 2  # Regular registration was called again
    assert cb2.exec_count == 1  # Unregistered by previous invoke, not called again
    assert cb3.exec_count == 1  # once=True, so not called again
    assert cb4.exec_count == 1  # Registered during previous invoke(), was called


# Timeout within 2 seconds
@pytest.mark.timeout(2)
@pytest.mark.flaky(reruns=3)
def test_random_port():

    # Starting port
    port = 9001
    # Test a set of continguous ports
    num_ports = 10
    # Number of times to try to find a port range
    n = 100

    # Make sure a single port can be found
    attempts = 0
    for _ in range(n):
        port_i = port - 1000 + attempts
        attempts += 1
        try:
            assert random_port(port_i, port_i) == port_i
            break
        except RuntimeError as e:
            print(e)
    if attempts == n:
        raise RuntimeError(f"Could not find a usable port in {n} tries")

    # Find a range of `num_ports` ports that are all available
    attempts = 0
    for _ in range(n):
        attempts += 1
        j = 0
        try:
            for j in range(num_ports):
                random_port(port + j, port + j)
            # If we reach this point, we have found a plausible range of ports to use
            break

        except RuntimeError as e:
            print(e)
            # Port `port + j` is busy,
            # Shift the test range and try again
            port += j + 1
            print("Trying port: ", port)
    # If no port is available, throw an error
    # `attempts` should be << n
    if attempts == n:
        raise RuntimeError(
            f"Could not find {num_ports} continguous ports to use for testing in {n} tries"
        )

    seen: Set[int] = set()
    # Ensure that `num_ports` unique random ports are eventually generated. If not (e.g. if the
    # max port number is treated as exclusive instead of inclusive, say) then the while
    # loop will not exit and the test will timeout.
    max_port = port + num_ports - 1
    while len(seen) < num_ports:
        seen.add(random_port(port, max_port))

    assert len(seen) == num_ports


def test_random_port_unusable():
    # 6000 is an unsafe port, make sure that it fails
    with pytest.raises(RuntimeError, match="Failed to find a usable random port"):
        random_port(6000, 6000)


def test_random_port_starvation():
    port = 9000
    for _ in range(100):
        try:
            with socketserver.TCPServer(
                ("127.0.0.1", port),
                socketserver.BaseRequestHandler,
            ):
                with pytest.raises(
                    RuntimeError, match="Failed to find a usable random port"
                ):
                    random_port(port, port)
        except OSError as e:
            print(e)
            # Port is busy, bump the port number
            port += 1
            print("Trying port: ", port)


def test_extract_js_keys():
    options = {
        "key1": "value1",
        "key2": js_eval("<h1>Hello, world!</h1>"),
        "key3": {
            "subkey1": js_eval("console.log('Hello, world!');"),
            "subkey2": "value2",
            "subkey3": {
                "subsubkey1": js_eval("<p>This is a paragraph.</p>"),
                "subsubkey2": "value3",
            },
        },
        "key4": "value4",
    }

    assert extract_js_keys(options) == [
        "key2",
        "key3.subkey1",
        "key3.subkey3.subsubkey1",
    ]



from shiny.express import input, render, ui

ui.page_opts(fillable=True)

with ui.card():
    with ui.layout_sidebar():
        with ui.sidebar(
            id="sidebar_left",
            open="desktop",
            title="Left sidebar",
            bg="dodgerBlue",
            class_="text-white",
            gap="20px",
            padding="10px",
            width="200px",
        ):
            "Left sidebar content"

        @render.code
        def state_left():
            return f"input.sidebar_left(): {input.sidebar_left()}"


with ui.card():
    with ui.layout_sidebar():
        with ui.sidebar(
            id="sidebar_right",
            position="right",
            open={"desktop": "closed", "mobile": "open"},
            padding=["10px", "20px"],
            bg="SlateBlue",
        ):
            "Right sidebar content"

        @render.code
        def state_right():
            return f"input.sidebar_right(): {input.sidebar_right()}"


with ui.card():
    with ui.layout_sidebar():
        with ui.sidebar(
            id="sidebar_closed",
            open="closed",
            bg="LightCoral",
            padding=["10px", "20px", "30px"],
        ):
            "Closed sidebar content"

        @render.code
        def state_closed():
            return f"input.sidebar_closed(): {input.sidebar_closed()}"


with ui.card():
    with ui.layout_sidebar():
        with ui.sidebar(
            id="sidebar_always",
            open="always",
            bg="PeachPuff",
            padding=["10px", "20px", "30px", "40px"],
            max_height_mobile="175px",
        ):
            "Always sidebar content"

        @render.code
        def state_always():
            return f"input.sidebar_always(): {input.sidebar_always()}"



from __future__ import annotations

import pandas as pd

from shiny.express import render, ui

df = pd.DataFrame(
    {
        "a": [1, 2, 3],
        "b": [4, 5, 6],
        "c": [7, 8, 9],
    }
)

ui.markdown(
    """
    * Edit a cell in column `b`.
    * Column `a` should be updated to `a value`
    * The edited cell should be back to a `ready` state
    """
)


@render.data_frame
def my_df():
    return render.DataTable(
        df,
        editable=True,
    )


# just testing if it works when editing another cell
@my_df.set_patches_fn
def edt(*, patches: tuple[render.CellPatch, ...]) -> list[render.CellPatch]:
    patch = patches[0]
    return [
        render.CellPatch(
            {
                "row_index": patch["row_index"],
                "column_index": 0,
                "value": "a value",
            },
        ),
    ]



# pyright: reportUnknownMemberType=false
from __future__ import annotations

import re
from copy import deepcopy
from typing import TYPE_CHECKING, Any, cast

from ..types import (
    Coordmap,
    CoordmapDims,
    CoordmapPanel,
    CoordmapPanelDomain,
    CoordmapPanelLog,
    CoordmapPanelMapping,
    CoordmapPanelRange,
    PlotnineFigure,
)

if TYPE_CHECKING:
    import numpy as np
    import numpy.typing as npt
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure
    from matplotlib.gridspec import SubplotSpec


def get_coordmap(fig: Figure) -> Coordmap | None:
    dims_ar = fig.get_size_inches() * fig.get_dpi()
    dims: CoordmapDims = {
        "width": dims_ar[0],
        "height": dims_ar[1],
    }

    all_axes = cast(
        "list[Axes]",  # pyright: ignore
        fig.get_axes(),
    )

    panels: list[CoordmapPanel] = []
    for i, axes in enumerate(all_axes):
        panel = get_coordmap_panel(axes, i + 1, dims["height"])
        panels.append(panel)

    coordmap: Coordmap = {
        "panels": panels,
        "dims": dims,
    }

    return coordmap


def get_coordmap_panel(axes: Axes, panel_num: int, height: float) -> CoordmapPanel:
    spspec = cast(
        "SubplotSpec",
        axes.get_subplotspec(),  # pyright: ignore[reportAttributeAccessIssue]
    )

    domain_xlim = axes.get_xlim()
    domain_ylim = axes.get_ylim()

    # Data coordinates of plotting area
    domain: CoordmapPanelDomain = {
        "left": domain_xlim[0],
        "right": domain_xlim[1],
        "bottom": domain_ylim[0],
        "top": domain_ylim[1],
    }

    # Pixel coordinates of plotting area
    transdata = axes.transData

    range_ar = cast(
        "npt.NDArray[np.double]",
        transdata.transform(
            [
                domain["left"],
                domain["bottom"],
                domain["right"],
                domain["top"],
            ]
        ),
    )

    # The values from transData.transform() have origin in the bottom-left, but we need
    # to provide coordinates with origin in upper-left.
    range: CoordmapPanelRange = {
        "left": range_ar[0],
        "right": range_ar[2],
        "bottom": height - range_ar[1],
        "top": height - range_ar[3],
    }

    log: CoordmapPanelLog = {"x": None, "y": None}
    xaxis: Any = axes.xaxis  # pyright: ignore[reportGeneralTypeIssues]
    if xaxis._scale.name == "log":
        log["x"] = xaxis._scale.base
        domain["left"] = xaxis._scale._transform.transform(domain["left"])
        domain["right"] = xaxis._scale._transform.transform(domain["right"])

    yaxis: Any = axes.yaxis  # pyright: ignore[reportGeneralTypeIssues]
    if yaxis._scale.name == "log":
        log["y"] = yaxis._scale.base
        domain["top"] = yaxis._scale._transform.transform(domain["top"])
        domain["bottom"] = yaxis._scale._transform.transform(domain["bottom"])

    return {
        "panel": panel_num,
        "row": spspec.rowspan.start + 1,
        "col": spspec.colspan.start + 1,
        # "panel_vars": {
        #     "panelvar1": "4",
        #     "panelvar2": "1",
        # },
        "domain": domain,
        "range": range,
        "log": log,
        "mapping": {
            "x": None,
            "y": None,
            # "x": "wt",
            # "y": "mpg",
            # "panelvar1": "cyl",
            # "panelvar2": "am",
        },
    }


def get_coordmap_plotnine(p: PlotnineFigure, fig: Figure) -> Coordmap | None:
    coordmap = get_coordmap(fig)

    if coordmap is None:
        return None

    p = deepcopy(p)
    p._build()  # pyright: ignore[reportAttributeAccessIssue]

    # Plotnine/ggplot figures can contain some information that is not in the matplotlib
    # Figure object that is generated.

    # The mappings are shared across all panels, so just get them once.
    mappings = _get_mappings(p)

    for i in range(len(coordmap["panels"])):
        # Copy the panel object; we'll mutate it, and then assign the copy back.
        panel = deepcopy(coordmap["panels"][i])
        panel_num = panel["panel"]

        panel["mapping"] = mappings.copy()

        # Slice out the row of the layout data frame that corresponds to this panel.
        layout_row = p.layout.layout.loc[p.layout.layout["PANEL"] == panel_num]

        # Get values of panelvars
        if "panelvar1" in panel["mapping"]:
            panel["panel_vars"] = {}
            panelvar1 = panel["mapping"]["panelvar1"]  # pyright: ignore
            # If panelvar1 is, say, "cyl", then panelvar1_val will be something like 4.
            panelvar1_val = _simplify_type(layout_row[panelvar1].iloc[0])
            panel["panel_vars"]["panelvar1"] = panelvar1_val  # pyright: ignore

        if "panelvar2" in panel["mapping"]:
            panelvar2 = panel["mapping"]["panelvar2"]  # pyright: ignore
            panelvar2_val = _simplify_type(layout_row[panelvar2].iloc[0])
            panel["panel_vars"]["panelvar2"] = panelvar2_val  # pyright: ignore

        # Get x and y scales
        xscale_num = layout_row["SCALE_X"].iloc[0]
        yscale_num = layout_row["SCALE_Y"].iloc[0]
        xscale = p.layout.panel_scales_x[xscale_num - 1]
        yscale = p.layout.panel_scales_y[yscale_num - 1]

        # Plotnine objects handle log scales a bit differently from regular matplotlib
        # Figures. Instead of using log scales in the matplotlib Figure object, it adds
        # log scales in the ggplot object.
        if hasattr(xscale, "_trans") and _is_log_trans(xscale._trans):
            panel["log"]["x"] = xscale._trans.base
        if hasattr(yscale, "_trans") and _is_log_trans(yscale._trans):
            panel["log"]["y"] = yscale._trans.base

        if hasattr(xscale, "_trans") and _is_reverse_trans(xscale._trans):
            domain = panel["domain"]
            panel["domain"]["left"] = -domain["left"]
            panel["domain"]["right"] = -domain["right"]
        if hasattr(yscale, "_trans") and _is_reverse_trans(yscale._trans):
            domain = panel["domain"]
            panel["domain"]["top"] = -domain["top"]
            panel["domain"]["bottom"] = -domain["bottom"]

        # Plotnine figures can also have transforms in the coordinates. We will assume
        # that log coord transforms are not used along with log scales, because that
        # would be really strange.
        if hasattr(p.layout.coord, "trans_x") and _is_log_trans(p.layout.coord.trans_x):
            panel["log"]["x"] = p.layout.coord.trans_x.base
        if hasattr(p.layout.coord, "trans_y") and _is_log_trans(p.layout.coord.trans_y):
            panel["log"]["y"] = p.layout.coord.trans_y.base

        # Assign temporary panel object back to coordmap.
        coordmap["panels"][i] = panel

    return coordmap


# Given a Plotnine transform object, report whether it is a log transform.
def _is_log_trans(trans: object) -> bool:
    return bool(re.fullmatch("log.*_trans", type(trans).__name__))


# Given a Plotnine transform object, report whether it is a reverse transform.
def _is_reverse_trans(trans: object) -> bool:
    return type(trans).__name__ == "reverse_trans"


def _get_mappings(p: PlotnineFigure) -> CoordmapPanelMapping:
    mapping: CoordmapPanelMapping = {"x": None, "y": None}

    if "x" in p.mapping:
        mapping["x"] = p.mapping["x"]
    if "y" in p.mapping:
        mapping["y"] = p.mapping["y"]

    if type(p.layout.coord).__name__ == "coord_flip":
        (mapping["x"], mapping["y"]) = (mapping["y"], mapping["x"])

    # The names (not values) of panel vars are the same across all panels.
    if type(p.layout.facet).__name__ == "facet_grid":
        n = 1
        if len(p.layout.facet.cols) > 0:
            mapping[f"panelvar{n}"] = p.layout.facet.cols[0]
            n += 1
        if len(p.layout.facet.rows) > 0:
            mapping[f"panelvar{n}"] = p.layout.facet.rows[0]

    elif type(p.layout.facet).__name__ == "facet_wrap":
        mapping["panelvar1"] = p.layout.facet.vars[0]

    return mapping


# Sometimes columns are types that are not JSON serializable, like np.int64. Convert
# these to types that are JSON serializable.
def _simplify_type(x: Any) -> Any:
    import numpy as np

    if isinstance(x, np.integer):
        return int(x)  # pyright: ignore[reportUnknownArgumentType]
    elif isinstance(x, np.floating):
        return float(x)  # pyright: ignore[reportUnknownArgumentType]
    else:
        return x



from shiny import App, Inputs, Outputs, Session, reactive, render, ui

app_ui = ui.page_sidebar(
    ui.sidebar("Sidebar content", id="sidebar"),
    ui.input_action_button("open_sidebar", label="Open sidebar", class_="me-3"),
    ui.input_action_button("close_sidebar", label="Close sidebar", class_="me-3"),
    ui.br(),
    ui.br(),
    ui.output_text_verbatim("state"),
    fillable=False,
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    @reactive.event(input.open_sidebar)
    def _():
        ui.update_sidebar("sidebar", show=True)

    @reactive.effect
    @reactive.event(input.close_sidebar)
    def _():
        ui.update_sidebar("sidebar", show=False)

    @render.text
    def state():
        return f"input.sidebar(): {input.sidebar()}"


app = App(app_ui, server=server)



from shiny.express import render
from shiny.types import SafeException


@render.ui
def safe():
    # This error _won't_ be sanitized when deployed
    raise SafeException("This is a safe exception")


@render.ui
def unsafe():
    # This error _will_ be sanitized when deployed
    raise Exception("This is an unsafe exception")



from __future__ import annotations

from contextlib import AbstractContextManager
from typing import Callable, TypeVar

from .._deprecated import warn_deprecated
from .._typing_extensions import ParamSpec
from ..render.renderer import RendererT
from .ui import hold

__all__ = ("suspend_display",)

P = ParamSpec("P")
R = TypeVar("R")
CallableT = TypeVar("CallableT", bound=Callable[..., object])


def output_args(
    **kwargs: object,
) -> Callable[[RendererT], RendererT]:
    """
    Sets default UI arguments for a Shiny rendering function.

    Each Shiny render function (like :class:`~shiny.render.plot`) can display itself when
    declared within a Shiny inline-style application. In the case of
    :class:`~shiny.render.plot`, the :func:`~shiny.ui.output_plot` function is called
    implicitly to display the plot. Use the `@ui_kwargs` decorator to specify arguments
    to be passed to `output_plot` (or whatever the corresponding UI function is) when
    the render function displays itself.

    Parameters
    ----------
    **kwargs
        Keyword arguments to be passed to the UI function.

    Returns
    -------
    :
        A decorator that sets the default UI arguments for a Shiny rendering function.
    """

    def wrapper(renderer: RendererT) -> RendererT:
        renderer._auto_output_ui_kwargs = kwargs
        return renderer

    return wrapper


def suspend_display(
    fn: Callable[P, R] | RendererT | None = None
) -> Callable[P, R] | RendererT | AbstractContextManager[None]:
    warn_deprecated(
        "`suspend_display` is deprecated. Please use `ui.hold` instead. "
        "It has a new name, but the exact same functionality."
    )
    return hold(fn)  # type: ignore



# pyright: reportUnnecessaryComparison=false

from __future__ import annotations

from ..types import Jsonifiable

__all__ = (
    "input_select",
    "input_selectize",
)

import copy
from json import dumps
from typing import Any, Mapping, Optional, Union, cast

from htmltools import Tag, TagChild, TagList, css, div, tags

from .._docstring import add_example
from .._namespaces import resolve_id
from ._html_deps_external import selectize_deps
from ._utils import JSEval, extract_js_keys, shiny_input_label

_Choices = Mapping[str, TagChild]
_OptGrpChoices = Mapping[str, _Choices]

# Canonical format for representing select options.
_SelectChoices = Union[_Choices, _OptGrpChoices]

# Formats available to the user
SelectChoicesArg = Union[
    # ["a", "b", "c"]
    "list[str]",
    # ("a", "b", "c")
    "tuple[str, ...]",
    # {"a": "Choice A", "b": tags.i("Choice B")}
    _Choices,
    # optgroup {"Group A": {"a1": "Choice A1", "a2": tags.i("Choice A2")}, "Group B": {}}
    _OptGrpChoices,
]


_topics = {
    "Server value": """
If `multiple=False`, the server value is a string with the value of the selected item.
If `multiple=True`, the server value is a tuple containing the values of the
selected items. When ``multiple=True`` and nothing is selected, this value
will be ``None``.
"""
}


@add_example()
def input_selectize(
    id: str,
    label: TagChild,
    choices: SelectChoicesArg,
    *,
    selected: Optional[str | list[str]] = None,
    multiple: bool = False,
    width: Optional[str] = None,
    remove_button: Optional[bool] = None,
    options: Optional[dict[str, Jsonifiable | JSEval]] = None,
) -> Tag:
    """
    Create a select list that can be used to choose a single or multiple items from a
    list of values.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    choices
        Either a list of choices or a dictionary mapping choice values to labels. Note
        that if a dictionary is provided, the keys are used as the (input) values so
        that the dictionary values can hold HTML labels. A dictionary of dictionaries is
        also supported, and in that case, the top-level keys are treated as
        ``<optgroup>`` labels.
    selected
        The values that should be initially selected, if any.
    multiple
        Is selection of multiple items allowed?
    width
        The CSS width, e.g. '400px', or '100%'
    remove_button
        Whether to add a remove button. This uses the `clear_button` and `remove_button`
        selectize plugins which can also be supplied as options. By default it will apply a
        remove button to multiple selections, but not single selections.
    options
        A dictionary of options. See the documentation of selectize.js for possible options.
        If you want to pass a JavaScript function, wrap the string in `ui.JS`.

    Returns
    -------
    :
        A UI element.

    Notes
    ------
    ::: {.callout-note title="Server value"}
    If `multiple=False`, the server value is a string with the value of the selected item.
    If `multiple=True`, the server value is a tuple containing the values of the
    selected items. When ``multiple=True`` and nothing is selected, this value
    will be ``None``.
    :::

    See Also
    --------
    * :func:`~shiny.ui.input_select`
    * :func:`~shiny.ui.input_radio_buttons`
    * :func:`~shiny.ui.input_checkbox_group`
    """

    x = input_select(
        id,
        label,
        choices,
        selected=selected,
        multiple=multiple,
        selectize=True,
        width=width,
        remove_button=remove_button,
        options=options,
    )

    return x


@add_example()
def input_select(
    id: str,
    label: TagChild,
    choices: SelectChoicesArg,
    *,
    selected: Optional[str | list[str]] = None,
    multiple: bool = False,
    selectize: bool = False,
    width: Optional[str] = None,
    size: Optional[str] = None,
    remove_button: Optional[bool] = None,
    options: Optional[dict[str, Jsonifiable | JSEval]] = None,
) -> Tag:
    """
    Create a select list that can be used to choose a single or multiple items from a
    list of values.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    choices
        Either a list of choices or a dictionary mapping choice values to labels. Note
        that if a dictionary is provided, the keys are used as the (input) values so
        that the dictionary values can hold HTML labels. A dictionary of dictionaries is
        also supported, and in that case, the top-level keys are treated as
        ``<optgroup>`` labels.
    selected
        The values that should be initially selected, if any.
    multiple
        Is selection of multiple items allowed?
    selectize
        Whether to use selectize.js or not.
    width
        The CSS width, e.g. '400px', or '100%'
    size
        Number of items to show in the selection box; a larger number will result in a
        taller box. Normally, when ``multiple=False``, a select input will be a
        drop-down list, but when size is set, it will be a box instead.

    Returns
    -------
    :
        A UI element.

    Notes
    ------
    ::: {.callout-note title="Server value"}
    If `multiple=False`, the server value is a string with the value of the selected item.
    If `multiple=True`, the server value is a tuple containing the values of the
    selected items. When ``multiple=True`` and nothing is selected, this value
    will be ``None``.
    :::

    See Also
    --------
    * :func:`~shiny.ui.input_selectize`
    * :func:`~shiny.ui.update_select`
    * :func:`~shiny.ui.input_radio_buttons`
    * :func:`~shiny.ui.input_checkbox_group`
    """
    if options is not None and selectize is False:
        raise Exception("Options can only be set when selectize is `True`.")

    remove_button = _resolve_remove_button(remove_button, multiple)

    choices_ = _normalize_choices(choices)
    if selected is None and not multiple:
        selected = _find_first_option(choices_)

    if options is None:
        options = {}

    opts = _update_options(options, remove_button, multiple)

    choices_tags = _render_choices(choices_, selected)

    resolved_id = resolve_id(id)

    return div(
        shiny_input_label(resolved_id, label),
        div(
            tags.select(
                *choices_tags,
                {"class": "shiny-input-select"},
                class_=None if selectize else "form-select",
                id=resolved_id,
                multiple=multiple,
                size=size,
            ),
            (
                TagList(
                    tags.script(
                        dumps(opts),
                        type="application/json",
                        data_for=resolved_id,
                        data_eval=dumps(extract_js_keys(opts)),
                    ),
                    selectize_deps(),
                )
                if selectize
                else None
            ),
        ),
        class_="form-group shiny-input-container",
        style=css(width=width),
    )


def _resolve_remove_button(remove_button: Optional[bool], multiple: bool) -> bool:
    if remove_button is None:
        if multiple:
            return True
        else:
            return False
    return remove_button


def _update_options(
    options: dict[str, Any], remove_button: bool, multiple: bool
) -> dict[str, Any]:
    opts = copy.deepcopy(options)
    plugins = opts.get("plugins", [])

    if remove_button:
        if multiple:
            to_add = "remove_button"
        else:
            to_add = "clear_button"

        if to_add not in plugins:
            plugins.append(to_add)

    if not plugins:
        return options

    opts["plugins"] = plugins
    return opts


def _normalize_choices(x: SelectChoicesArg) -> _SelectChoices:
    if x is None:
        raise TypeError("`choices` must be a list, tuple, or dict.")
    elif isinstance(x, (list, tuple)):
        return {k: k for k in x}
    else:
        return x


def _render_choices(
    x: _SelectChoices, selected: Optional[str | list[str]] = None
) -> TagList:
    result = TagList()

    if x is None:
        return result

    for k, v in x.items():
        if isinstance(v, Mapping):
            result.append(
                tags.optgroup(
                    *(_render_choices(cast(_SelectChoices, v), selected)), label=k
                )
            )
        else:
            is_selected = False
            if isinstance(selected, list):
                is_selected = k in selected
            else:
                is_selected = k == selected

            result.append(tags.option(v, value=k, selected=is_selected))

    return result


# Returns the first option in a _SelectChoices object. For most cases, this is
# straightforward. In the following, the first option is "a":
# {"a": "Choice A", "b": "Choice B", "c": "Choice C"}
#
# Sometimes the first option is nested within an optgroup. For example, in the
# following, the first option is "b1":
# {
#     "Group A": {},
#     "Group B": {"Choice B1": "b1", "Choice B2": "b2"},
# }
def _find_first_option(x: _SelectChoices) -> Optional[str]:
    if x is None:
        return None

    for k, v in x.items():
        if isinstance(v, dict):
            result = _find_first_option(cast(_SelectChoices, v))
            if result is not None:
                return result
        else:
            return k

    return None



import re

import pytest
from playwright.sync_api import Page
from utils.deploy_utils import skip_if_not_chrome

from shiny.playwright import controller
from shiny.render._data_frame import ColumnFilter
from shiny.run import ShinyAppProc


@skip_if_not_chrome
@pytest.mark.parametrize("df_type", ["pandas", "polars"])
def test_validate_html_columns(
    page: Page, local_app: ShinyAppProc, df_type: str
) -> None:
    page.goto(local_app.url)

    data_frame = controller.OutputDataFrame(page, f"{df_type}_df")

    tab = controller.NavsetCardUnderline(page, "tab")
    output_txt = controller.OutputTextVerbatim(page, f"{df_type}_test_cell_text")
    test_button = controller.InputActionButton(page, f"{df_type}_test_cell_button")

    tab.set(df_type)

    # verify shiny reactive output UI in cell
    # Add a larger timeout for CI
    output_txt.expect_value(f"{df_type}_test_cell_value 0", timeout=30 * 1000)

    # Test Shiny reactive output in cell
    test_button.click()
    output_txt.expect_value(f"{df_type}_test_cell_value 1")

    # Assert cell content is not "null", but `""`
    data_frame.expect_cell(re.compile(r"^$"), row=3, col=13)
    data_frame.expect_cell(re.compile(r"^$"), row=1, col=16)

    # assert patching works
    data_frame.expect_cell("N1A1", row=0, col=6)
    data_frame.set_cell("N1A11", row=0, col=6, finish_key="Enter")
    data_frame.expect_cell("ID: N1A11", row=0, col=6)

    # assert sorting works
    data_frame.expect_cell("1", row=0, col=1)
    data_frame.set_sort(1)  # sample
    data_frame.expect_cell("152", row=0, col=1)
    # confirm sorting is reset before setting
    data_frame.set_sort(1)  # sample
    data_frame.expect_cell("152", row=0, col=1)
    # confirm sorting is reset before setting
    data_frame.set_sort(13)  # sex
    data_frame.expect_cell("4", row=0, col=1)

    # sort with multiple columns
    data_frame.set_sort([8, 1])  # Date egg, sample
    data_frame.expect_cell("34", row=0, col=1)
    # Perform double sort, make sure extra clicks are made to sort the second sort column
    data_frame.set_sort([8, 9])  # Dat egg, Culmen Length (mm)
    data_frame.expect_cell("10", row=0, col=1)
    data_frame.set_sort([8, {"col": 9, "desc": False}])
    data_frame.expect_cell("9", row=0, col=1)
    # reset sorting
    data_frame.set_sort(None)
    data_frame.expect_cell("1", row=0, col=1)
    # set sorting for rest of tests
    data_frame.set_sort({"col": 1, "desc": True})
    data_frame.expect_cell("152", row=0, col=1)

    # if a column is sorted, editing should not change the order
    data_frame.set_cell("152", row=0, col=1, finish_key="Enter")
    data_frame.expect_cell("151", row=1, col=1)

    # assert HTMLDependency works by verifying javascript variable
    test_value = page.evaluate("window.shinytestvalue")
    assert test_value == "testing"

    # Sorting should not work for columns that are HTML columns
    # Verify that the sorting is reset
    with pytest.raises(AssertionError) as e:
        data_frame.set_sort(3, timeout=0)
        assert "header-html" in str(e)
    data_frame.expect_cell("1", row=0, col=1)

    # filter by Individual IDs
    filter_text: ColumnFilter = {"col": 6, "value": "N2"}
    data_frame.set_filter(filter_text)
    data_frame.expect_cell("3", row=0, col=1)

    # respect filtering even after edits when filters have been applied
    data_frame.set_cell("3", row=0, col=1, finish_key="Enter")
    data_frame.expect_cell("4", row=1, col=1)

    # assert that html columns are not editable
    data_frame.expect_cell_class("cell-html", row=1, col=0)
    data_frame.expect_cell_class("cell-html", row=0, col=2)
    data_frame.expect_cell_class("cell-html", row=0, col=3)
    data_frame.expect_cell_class("cell-html", row=0, col=4)
    data_frame.expect_cell_class("cell-html", row=0, col=0)

    data_frame.cell_locator(row=0, col=0).click()
    # Verify the class does not change to editing when a cell under a HTML column is clicked
    data_frame.expect_cell_class("cell-html", row=0, col=0)

    # Filter using a range for a column that contains numbers
    filter_num_range: list[ColumnFilter] = [
        {"col": 6, "value": "N2"},
        {"col": 1, "value": (40, 50)},
    ]
    data_frame.set_filter(filter_num_range)
    data_frame.expect_cell("40", row=0, col=1)

    # Editing a cell in the first row and hitting `shift+enter` should not submit the change and stay editing the current cell
    data_frame.expect_cell("N25A2", row=0, col=6)
    data_frame.set_cell("NAAAAA", row=0, col=6, finish_key="Shift+Enter")
    data_frame.expect_cell("N25A2", row=0, col=6)
    data_frame.set_cell("NBBBBB", row=0, col=6, finish_key="Escape")
    data_frame.expect_cell("N25A2", row=0, col=6)

    # Editing a cell in the last row and hitting `enter` should not submit the change and stay editing the current cell
    # data_frame.set_column_filter(7, text="No")
    # Test scrolling to last row
    data_frame.set_cell("NAAAAA", row=7, col=6, finish_key="Enter")
    data_frame.expect_cell("N29A2", row=7, col=6)
    data_frame.set_cell("NAAAAA", row=7, col=6, finish_key="Escape")
    data_frame.expect_cell("N29A2", row=7, col=6)

    # Test scrolling up to top
    data_frame.expect_cell("N25A2", row=0, col=6)



from shiny.express import render, ui

with ui.navset_card_tab(id="tab"):

    with ui.nav_panel("Tab 1"):
        "Tab 1 content"
    with ui.nav_panel("Text Area"):
        ui.input_text_area(
            id="test_text_area",
            label="A text area input",
            autoresize=True,
            value="a\nb\nc\nd\ne",
        )

        ui.input_text_area(
            id="test_text_area2",
            label="A second text area input",
            autoresize=True,
            value="a\nb\nc\nd\ne",
            rows=4,
        )


@render.code
def text():
    return "Loaded"



from shiny import reactive
from shiny.express import input, ui

ui.markdown("The checkbox group controls the select input")

ui.input_checkbox_group(
    "inCheckboxGroup", "Input checkbox", ["Item A", "Item B", "Item C"]
)
ui.input_select("inSelect", "Select input", ["Item A", "Item B", "Item C"])


@reactive.effect
def _():
    x = input.inCheckboxGroup()

    # Can use [] to remove all choices
    if x is None:
        x = []
    elif isinstance(x, str):
        x = [x]

    ui.update_select(
        "inSelect",
        label="Select input label " + str(len(x)),
        choices=x,
        selected=x[len(x) - 1] if len(x) > 0 else None,
    )



import pandas  # noqa: F401 (this line needed for Shinylive to load plotly.express)
import plotly.express as px
from shinywidgets import output_widget, render_widget

from shiny import App, reactive, render, req, ui

# Load the Gapminder dataset
df = px.data.gapminder()

# Prepare a summary DataFrame
summary_df = (
    df.groupby("country")
    .agg(
        {
            "pop": ["min", "max", "mean"],
            "lifeExp": ["min", "max", "mean"],
            "gdpPercap": ["min", "max", "mean"],
        }
    )
    .reset_index()
)

summary_df.columns = ["_".join(col).strip() for col in summary_df.columns.values]
summary_df.rename(columns={"country_": "country"}, inplace=True)

app_ui = ui.page_fillable(
    {"class": "p-3"},
    ui.markdown(
        "**Instructions**: Select one or more countries in the table below to see more information."
    ),
    ui.layout_columns(
        ui.card(ui.output_data_frame("summary_data"), height="400px"),
        ui.card(output_widget("country_detail_pop"), height="400px"),
        ui.card(output_widget("country_detail_percap"), height="400px"),
        col_widths=[12, 6, 6],
    ),
)


def server(input, output, session):
    @render.data_frame
    def summary_data():
        return render.DataGrid(summary_df.round(2), selection_mode="rows")

    @reactive.calc
    def filtered_df():
        data_selected = summary_data.data_view(selected=True)
        req(not data_selected.empty)
        countries = data_selected["country"]
        # Filter data for selected countries
        return df[df["country"].isin(countries)]

    @render_widget
    def country_detail_pop():
        return px.line(
            filtered_df(),
            x="year",
            y="pop",
            color="country",
            title="Population Over Time",
        )

    @render_widget
    def country_detail_percap():
        return px.line(
            filtered_df(),
            x="year",
            y="gdpPercap",
            color="country",
            title="GDP per Capita Over Time",
        )


app = App(app_ui, server)



import pytest
from playwright.sync_api import Page
from utils.deploy_utils import (
    local_deploys_app_url_fixture,
    reruns,
    reruns_delay,
    skip_if_not_chrome,
)

from shiny.playwright import controller

app_url = local_deploys_app_url_fixture("shiny_express_accordion")


@skip_if_not_chrome
@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
def test_express_accordion(page: Page, app_url: str) -> None:
    page.goto(app_url)

    acc = controller.Accordion(page, "express_accordion")
    acc_panel_2 = acc.accordion_panel("Panel 2")
    acc_panel_2.expect_open(True)
    acc_panel_2.expect_body("n = 50")
    acc_panel_2.set(False)
    acc_panel_2.expect_open(False)



from datetime import date, timedelta

from shiny import reactive
from shiny.express import input, ui

ui.input_slider("n", "Day of month", min=1, max=30, value=10)
ui.input_date_range("inDateRange", "Input date")


@reactive.effect
def _():
    d = date(2013, 4, input.n())
    ui.update_date_range(
        "inDateRange",
        label="Date range label " + str(input.n()),
        start=d - timedelta(days=1),
        end=d + timedelta(days=1),
        min=d - timedelta(days=5),
        max=d + timedelta(days=5),
    )



import random
import time

from shiny import reactive
from shiny.express import input, render, ui


@reactive.calc
def first():
    input.first()
    p = ui.Progress()
    for i in range(30):
        p.set(i / 30, message="Computing, please wait...")
        time.sleep(0.1)
    p.close()
    return random.randint(1, 1000)


@reactive.calc
def second():
    input.second()
    return random.randint(1, 1000)


with ui.card():
    with ui.layout_columns():
        ui.input_action_button("first", "Invalidate first (slow) computation")
        ui.input_action_button("second", "Invalidate second (fast) computation")

    @render.text
    def result():
        return first() + second()



from shiny import req
from shiny.express import input, render, ui

ui.input_checkbox_group(
    "colors",
    "Choose color(s):",
    {
        "red": ui.span("Red", style="color: #FF0000;"),
        "green": ui.span("Green", style="color: #00AA00;"),
        "blue": ui.span("Blue", style="color: #0000AA;"),
    },
)


@render.ui
def val():
    req(input.colors())
    return "You chose " + ", ".join(input.colors())



import faicons as fa
import plotly.express as px

# Load data and compute static values
from shared import app_dir, tips
from shinywidgets import render_plotly

from shiny import reactive, render
from shiny.express import input, ui

bill_rng = (min(tips.total_bill), max(tips.total_bill))

# Add page title and sidebar
ui.page_opts(title="Restaurant tipping", fillable=True)

with ui.sidebar(open="desktop"):
    ui.input_slider(
        "total_bill",
        "Bill amount",
        min=bill_rng[0],
        max=bill_rng[1],
        value=bill_rng,
        pre="$",
    )
    ui.input_checkbox_group(
        "time",
        "Food service",
        ["Lunch", "Dinner"],
        selected=["Lunch", "Dinner"],
        inline=True,
    )
    ui.input_action_button("reset", "Reset filter")

# Add main content
ICONS = {
    "user": fa.icon_svg("user", "regular"),
    "wallet": fa.icon_svg("wallet"),
    "currency-dollar": fa.icon_svg("dollar-sign"),
    "ellipsis": fa.icon_svg("ellipsis"),
}

with ui.layout_columns(fill=False):
    with ui.value_box(showcase=ICONS["user"]):
        "Total tippers"

        @render.express
        def total_tippers():
            tips_data().shape[0]

    with ui.value_box(showcase=ICONS["wallet"]):
        "Average tip"

        @render.express
        def average_tip():
            d = tips_data()
            if d.shape[0] > 0:
                perc = d.tip / d.total_bill
                f"{perc.mean():.1%}"

    with ui.value_box(showcase=ICONS["currency-dollar"]):
        "Average bill"

        @render.express
        def average_bill():
            d = tips_data()
            if d.shape[0] > 0:
                bill = d.total_bill.mean()
                f"${bill:.2f}"


with ui.layout_columns(col_widths=[6, 6, 12]):
    with ui.card(full_screen=True):
        ui.card_header("Tips data")

        @render.data_frame
        def table():
            return render.DataGrid(tips_data())

    with ui.card(full_screen=True):
        with ui.card_header(class_="d-flex justify-content-between align-items-center"):
            "Total bill vs tip"
            with ui.popover(title="Add a color variable", placement="top"):
                ICONS["ellipsis"]
                ui.input_radio_buttons(
                    "scatter_color",
                    None,
                    ["none", "sex", "smoker", "day", "time"],
                    inline=True,
                )

        @render_plotly
        def scatterplot():
            color = input.scatter_color()
            return px.scatter(
                tips_data(),
                x="total_bill",
                y="tip",
                color=None if color == "none" else color,
                trendline="lowess",
            )

    with ui.card(full_screen=True):
        with ui.card_header(class_="d-flex justify-content-between align-items-center"):
            "Tip percentages"
            with ui.popover(title="Add a color variable"):
                ICONS["ellipsis"]
                ui.input_radio_buttons(
                    "tip_perc_y",
                    "Split by:",
                    ["sex", "smoker", "day", "time"],
                    selected="day",
                    inline=True,
                )

        @render_plotly
        def tip_perc():
            from ridgeplot import ridgeplot

            dat = tips_data()
            dat["percent"] = dat.tip / dat.total_bill
            yvar = input.tip_perc_y()
            uvals = dat[yvar].unique()

            samples = [[dat.percent[dat[yvar] == val]] for val in uvals]

            plt = ridgeplot(
                samples=samples,
                labels=uvals,
                bandwidth=0.01,
                colorscale="viridis",
                colormode="row-index",
            )

            plt.update_layout(
                legend=dict(
                    orientation="h", yanchor="bottom", y=1.02, xanchor="center", x=0.5
                )
            )

            return plt


ui.include_css(app_dir / "styles.css")

# --------------------------------------------------------
# Reactive calculations and effects
# --------------------------------------------------------


@reactive.calc
def tips_data():
    bill = input.total_bill()
    idx1 = tips.total_bill.between(bill[0], bill[1])
    idx2 = tips.time.isin(input.time())
    return tips[idx1 & idx2]


@reactive.effect
@reactive.event(input.reset)
def _():
    ui.update_slider("total_bill", value=bill_rng)
    ui.update_checkbox_group("time", selected=["Lunch", "Dinner"])



from __future__ import annotations

from enum import Enum
from typing import Dict, Iterable, Literal, Optional, Tuple, TypeVar, Union, cast
from warnings import warn

from htmltools import Tag, TagAttrs, TagAttrValue, TagChild, css

from .._docstring import add_example
from ._html_deps_shinyverse import components_dependencies
from ._layout import wrap_all_in_gap_spaced_container
from ._tag import consolidate_attrs
from .css import CssUnit, as_css_unit
from .fill import as_fill_item

T = TypeVar("T")


Breakpoint = Literal["xs", "sm", "md", "lg", "xl", "xxl"]
"""
References
----------
* [Available Bootstrap breakpoints](https://getbootstrap.com/docs/5.3/layout/breakpoints/#available-breakpoints)
"""


breakpoints: Tuple[Breakpoint, ...] = ("xs", "sm", "md", "lg", "xl", "xxl")


BreakpointsSoft = Dict[Breakpoint, Union[Iterable[T], T, None]]
BreakpointsOptional = Dict[Breakpoint, Union[Iterable[T], None]]
BreakpointsComplete = Dict[Breakpoint, Iterable[T]]
BreakpointsUser = Union[BreakpointsSoft[T], Iterable[T], T, None]


@add_example()
def layout_columns(
    *args: TagChild | TagAttrs,
    col_widths: BreakpointsUser[int] = None,
    row_heights: BreakpointsUser[CssUnit] = None,
    fill: bool = True,
    fillable: bool = True,
    gap: Optional[CssUnit] = None,
    class_: Optional[str] = None,
    height: Optional[CssUnit] = None,
    min_height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Create responsive, column-based grid layouts, based on a 12-column grid.

    Parameters
    ----------
    *args
        Child elements or attributes to be added to the layout.
    col_widths
        The widths of the columns, possibly at different breakpoints. Can be one of the
        following:

        * `None` (the default): Automatically determines a sensible number of columns
          based on the number of children given to the layout.
        * A list or tuple of integers between 1 and 12, where each element represents
          the number of columns for the relevant UI element. Column widths are recycled
          to extend the values in `col_widths` to match the actual number of items in
          the layout, and children are wrapped onto the next row when a row exceeds 12
          column units. For example, `col_widths=(4, 8, 12)` allocates 4 columns to the
          first element, 8 columns to the second element, and 12 columns to the third
          element (which wraps to the next row). Negative values are also allowed, and
          are treated as empty columns. For example, `col_widths=(-2, 8, -2)` would
          allocate 8 columns to an element (with 2 empty columns on either side).
        * A dictionary of column widths at different breakpoints. The keys should be
          one of `"xs"`, `"sm"`, `"md"`, `"lg"`, `"xl"`, or `"xxl"`, and the values are
          either of the above. For example, `col_widths={"sm": (3, 3, 6), "lg": (4)}`.
    row_heights
        The heights of the rows, possibly at different breakpoints. Can be one of the
        following:

        * A numeric vector, where each value represents the
          [fractional unit](https://css-tricks.com/introduction-fr-css-unit/)
          (`fr`) height of the relevant row. If there are more rows than values
          provided, the pattern will be repeated. For example, `row_heights=(1, 2)`
          allows even rows to take up twice as much space as odd rows.
        * A list of numeric or CSS length units, where each value represents the height
          of the relevant row. If more rows are needed than values provided, the pattern
          will repeat. For example, `row_heights=["auto", 1]` allows the height of odd
          rows to be driven my it's contents and even rows to be
          [`1fr`](https://css-tricks.com/introduction-fr-css-unit/).
        * A single string containing CSS length units. In this case, the value is
          supplied directly to `grid-auto-rows`.
        * A dictionary of row heights at different breakpoints, where each key is a
          breakpoint name (one of `"xs"`, `"sm"`, `"md"`, `"lg"`, `"xl"`, or `"xxl"`)
          and where the values may be any of the above options.
    fill
        Whether or not to allow the layout to grow/shrink to fit a fillable container
        with an opinionated height (e.g., :func:`~shiny.ui.page_fillable`).
    fillable
        Whether or not each element is wrapped in a fillable container.
    gap
        Any valid CSS unit to use for the gap between columns.
    class_
        CSS class(es) to apply to the containing element.
    height,max_height,min_height
        A valid CSS unit (e.g., `height="200px"`). Use `min_height` and `max_height` in
        a filling layout to ensure that the layout container does not shrink below a
        `min_height` or grow beyond a `max_height`.
    **kwargs
        Additional attributes to apply to the containing element.

    Returns
    -------
    :
        An :class:`~htmltools.Tag` element.

    See Also
    --------
    * :func:`~shiny.ui.layout_column_wrap` for laying out elements into a uniform grid.

    Reference
    --------
    * [Bootstrap CSS Grid](https://getbootstrap.com/docs/5.3/layout/grid/)
    * [Bootstrap Breakpoints](https://getbootstrap.com/docs/5.3/layout/breakpoints/)
    """
    attrs, children = consolidate_attrs(*args, class_=class_, **kwargs)

    col_widths_spec = as_col_spec(col_widths, len(children))

    # Create the bslib-layout-columns element
    tag = Tag(
        "bslib-layout-columns",
        {
            "class": "bslib-grid grid bslib-mb-spacing",
            "style": css(
                gap=as_css_unit(gap),
                height=as_css_unit(height),
                min_height=as_css_unit(min_height),
                max_height=as_css_unit(max_height),
            ),
        },
        col_widths_attrs(col_widths_spec),
        attrs,
        row_heights_attrs(row_heights),
        *wrap_all_in_gap_spaced_container(children, fillable, "bslib-grid-item"),
        components_dependencies(),
    )

    # Apply fill to the outer layout (fillable is applied to the children)
    if fill:
        tag = as_fill_item(tag)

    return tag


def as_col_spec(
    col_widths: BreakpointsUser[int],
    n_kids: int,
) -> BreakpointsOptional[int] | None:
    if col_widths is None:
        return None

    if not isinstance(col_widths, Dict):
        return {"sm": validate_col_width(col_widths, n_kids, "sm")}

    ret: BreakpointsOptional[int] = {}
    col_widths_items = cast(BreakpointsSoft[int], col_widths).items()

    for brk, value in col_widths_items:
        if brk not in breakpoints:
            raise ValueError(
                f"Breakpoint '{brk}' is not valid. Valid breakpoints are: {', '.join(breakpoints)}'."
            )

        if value is None:
            ret[brk] = None
        elif isinstance(value, (int, Iterable)):
            ret[brk] = validate_col_width(value, n_kids, brk)
        else:
            raise TypeError(
                f"Invalid type for value at breakpoint '{brk}'. Expected int or Iterable[int]."
            )

    return ret


def validate_col_width(
    x: Iterable[int] | int, n_kids: int, break_name: Breakpoint
) -> Iterable[int]:
    if isinstance(x, int):
        y = [x]
    else:
        y = x

    if not all(isinstance(i, int) for i in y):
        raise ValueError(
            f"Column values at breakpoint '{break_name}' must be integers. Values greater than 0 indicate width, and negative values indicate a column offset."
        )

    if any(i == 0 for i in y):
        raise ValueError(
            f"Column values at breakpoint '{break_name}' must be greater than 0 to indicate width, or negative to indicate a column offset."
        )

    if not any(b > 0 for b in y):
        raise ValueError(
            f"Column values at breakpoint '{break_name}' must include at least one positive integer width."
        )

    if len(list(y)) > n_kids:
        warn(
            f"More column widths than children at breakpoint '{break_name}', extra widths will be ignored."
        )

    return y


def col_widths_attrs(col_widths: BreakpointsOptional[int] | None) -> TagAttrs:
    ret: Dict[str, TagAttrValue] = {}
    if col_widths is None:
        return ret

    for break_name, value in col_widths.items():
        if isinstance(break_name, Enum):
            break_name = break_name.value
        break_name = f"col-widths-{break_name}"
        if value is None:
            ret[break_name] = ""
        else:
            ret[break_name] = ",".join([str(v) for v in value])

    return ret


def maybe_fr_unit(x: CssUnit) -> str:
    if isinstance(x, float):
        x = round(x)

    if isinstance(x, int):
        return f"{x}fr"

    return x


def row_heights_attrs(
    x: BreakpointsUser[CssUnit],
) -> TagAttrs:
    if x is None:
        return {"style": "", "class": ""}

    if isinstance(x, (int, float, str)):  # CssUnit
        x = [x]

    if hasattr(x, "__iter__") and not isinstance(x, Dict):
        # For a single row_heights, we use the same value across all breakpoints,
        # including mobile
        height = " ".join([maybe_fr_unit(h) for h in x])
        return {
            "style": css(
                **{"--bslib-grid--row-heights": height},
            ),
            "class": "",
        }

    x = cast(BreakpointsSoft[CssUnit], x)

    # Remove any None values from x
    x_complete = {k: v for k, v in x.items() if v is not None}

    # We use classes to activate CSS variables at the right breakpoints.
    classes = [f"bslib-grid--row-heights--{brk}" for brk in x_complete.keys()]

    # Create CSS variables, treating numeric values as fractional units, passing strings
    css_vars: Dict[str, str] = {}
    for brk, heights in x_complete.items():
        var = f"--bslib-grid--row-heights--{brk}"

        # We'd use isinstance(heights, CssUnit) (needs Python > 3.9)
        # or isinstance_cssunit(heights) (pyright 1.1.369+ doesn't like it)
        if isinstance(heights, (str, int, float)):
            heights = [heights]

        value = " ".join([maybe_fr_unit(h) for h in heights])
        css_vars[var] = value

    return {
        "style": css(**css_vars),
        "class": " ".join(classes),
    }



from __future__ import annotations

from copy import copy
from typing import TypeVar

from htmltools import Tag, TagAttrs

from ..._docstring import add_example, no_example
from .._html_deps_shinyverse import fill_dependency

__all__ = (
    "as_fillable_container",
    "as_fill_item",
    "remove_all_fill",
    # "is_fill_item",
    # "is_fillable_container",
)

TagT = TypeVar("TagT", bound="Tag")


FILL_ITEM_CLASS = "html-fill-item"
FILL_CONTAINER_CLASS = "html-fill-container"

# Attributes which must be used as a prepended value.
# This allows for the user to add a class and overright the value.
# THIS OBJECTS SHOULD BE USED ONLY IF NECESSARY!; Ex: `page_fillable()` should send
# *arg/**kwargs to `page_bootstrap()` rather than nesting alterted tags.body() objects.
FILL_ITEM_ATTRS: TagAttrs = {"class": FILL_ITEM_CLASS}
FILLABLE_CONTAINTER_ATTRS: TagAttrs = {"class": FILL_CONTAINER_CLASS}


@add_example()
def as_fillable_container(
    tag: TagT,
) -> TagT:
    """
    Coerce a tag to a fillable container.

    Filling layouts are built on the foundation of _fillable containers_ and _fill
    items_ (_fill carriers_ are both _fillable containers_ and _fill items_). This is
    why most UI components (e.g., :func:`~shiny.ui.card`,
    :func:`~shiny.ui.layout_sidebar`) possess both `fillable` and `fill` arguments (to
    control their fill behavior). However, sometimes it's useful to add, remove, and/or
    test fillable/fill properties on arbitrary :class:`~htmltools.Tag`, which these
    functions are designed to do.

    Parameters
    ----------
    tag
        a Tag object.

    Returns
    -------
    :
        A copy of the original :class:`~htmltools.Tag` object (`tag`) with additional
        attributes (and an :class:`~htmltools.HTMLDependency`).

    See Also
    --------
    * :func:`~shiny.ui.fill.as_fill_item`
    * :func:`~shiny.ui.fill.remove_all_fill`
    """
    res = copy(tag)
    res.add_class(FILL_CONTAINER_CLASS)
    res.append(fill_dependency())
    return res


@add_example()
def as_fill_item(
    tag: TagT,
) -> TagT:
    """
    Coerce a tag to a fill item.

    Filling layouts are built on the foundation of _fillable containers_ and _fill
    items_ (_fill carriers_ are both _fillable containers_ and _fill items_). This is
    why most UI components (e.g., :func:`~shiny.ui.card`,
    :func:`~shiny.ui.layout_sidebar`) possess both `fillable` and `fill` arguments (to
    control their fill behavior). However, sometimes it's useful to add, remove, and/or
    test fillable/fill properties on arbitrary :class:`~htmltools.Tag`, which these
    functions are designed to do.

    Parameters
    ----------
    tag
        a Tag object.

    Returns
    -------
    :
        A copy of the original :class:`~htmltools.Tag` object (`tag`) with additional
        attributes (and an :class:`~htmltools.HTMLDependency`).

    See Also
    --------
    * :func:`~shiny.ui.fill.as_fillable_container`
    * :func:`~shiny.ui.fill.remove_all_fill`
    """
    res = copy(tag)
    res.add_class(FILL_ITEM_CLASS)
    res.append(fill_dependency())
    return res


@no_example()
def remove_all_fill(
    tag: TagT,
) -> TagT:
    """
    Remove any filling layouts from a tag.

    Filling layouts are built on the foundation of _fillable containers_ and _fill
    items_ (_fill carriers_ are both _fillable containers_ and _fill items_). This is
    why most UI components (e.g., :func:`~shiny.ui.card`,
    :func:`~shiny.ui.layout_sidebar`) possess both `fillable` and `fill` arguments (to
    control their fill behavior). However, sometimes it's useful to add, remove, and/or
    test fillable/fill properties on arbitrary :class:`~htmltools.Tag`, which these
    functions are designed to do.

    Parameters
    ----------
    tag
        a Tag object.

    Returns
    -------
    :
        A copy of the original :class:`~htmltools.Tag` object with filling layout
        attributes removed.


    See Also
    --------
    * :func:`~shiny.ui.fill.as_fill_item`
    * :func:`~shiny.ui.fill.as_fillable_container`
    """

    res = copy(tag)
    res.remove_class(FILL_CONTAINER_CLASS)
    res.remove_class(FILL_ITEM_CLASS)
    return res


# Method currently not exposed, but implemented within bslib
def is_fillable_container(
    tag: object,
) -> bool:
    """
    Test a tag for being a fillable container

    Filling layouts are built on the foundation of _fillable containers_ and _fill
    items_ (_fill carriers_ are both _fillable containers_ and _fill items_). This is
    why most UI components (e.g., :func:`~shiny.ui.card` and
    :func:`~shiny.ui.layout_sidebar`) possess both `fillable` and `fill` arguments (to
    control their fill behavior). However, sometimes it's useful to add, remove, and/or
    test fillable/fill properties on arbitrary :class:`~htmltools.Tag`, which these
    functions are designed to do.

    Parameters
    ----------
    tag
        a Tag object.

    Returns
    -------
    :
        Whether or not `tag` is a fillable container.


    See Also
    --------
    * :func:`~shiny.ui.fill.as_fill_item`
    * :func:`~shiny.ui.fill.as_fillable_container`
    * :func:`~shiny.ui.fill.remove_all_fill`
    """
    # TODO-future; Handle widgets
    # # won't actually work until (htmltools#334) gets fixed
    # renders_to_tag_class(x, FILL_CONTAINER_CLASS, ".html-widget")

    return isinstance(tag, Tag) and tag.has_class(FILL_CONTAINER_CLASS)


# Method currently not exposed, but implemented within bslib
def is_fill_item(tag: object) -> bool:
    """
    Test a tag for being a fill item

    Filling layouts are built on the foundation of _fillable containers_ and _fill
    items_ (_fill carriers_ are both _fillable containers_ and _fill items_). This is
    why most UI components (e.g., :func:`~shiny.ui.card` and
    :func:`~shiny.ui.layout_sidebar`) possess both `fillable` and `fill` arguments (to
    control their fill behavior). However, sometimes it's useful to add, remove, and/or
    test fillable/fill properties on arbitrary :class:`~htmltools.Tag`, which these
    functions are designed to do.

    Parameters
    ----------
    tag
        a Tag object.

    Returns
    -------
    :
        Whether or not `tag` is a fill item.

    See Also
    --------
    * :func:`~shiny.ui.fill.as_fill_item`
    * :func:`~shiny.ui.fill.as_fillable_container`
    * :func:`~shiny.ui.fill.remove_all_fill`
    """
    # TODO-future; Handle widgets
    # # won't actually work until (htmltools#334) gets fixed
    # renders_to_tag_class(x, FILL_ITEM_CLASS, ".html-widget")

    return isinstance(tag, Tag) and tag.has_class(FILL_ITEM_CLASS)



from shiny import reactive
from shiny.express import input, ui

ui.markdown("The first radio button group controls the second")

ui.input_radio_buttons(
    "inRadioButtons", "Input radio buttons", ["Item A", "Item B", "Item C"]
)
ui.input_radio_buttons(
    "inRadioButtons2", "Input radio buttons 2", ["Item A", "Item B", "Item C"]
)


@reactive.effect
def _():
    x = input.inRadioButtons()

    # Can also set the label and select items
    ui.update_radio_buttons(
        "inRadioButtons2",
        label="Radio buttons label " + x,
        choices=[x],
        selected=x,
    )



import asyncio
import hashlib
import time

from shiny import App, Inputs, Outputs, Session, reactive, render, ui


def calc(value: str) -> str:
    # Simluate this taking a long time
    time.sleep(1)
    m = hashlib.sha256()
    m.update(value.encode("utf-8"))
    return m.digest().hex()


app_ui = ui.page_fluid(
    ui.input_text_area(
        "value", "Value to sha256sum", value="The quick brown fox", rows=5, width="100%"
    ),
    ui.p(ui.input_action_button("go", "Calculate"), class_="mb-3"),
    ui.output_text_verbatim("hash_output"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.text()
    @reactive.event(input.go)
    async def hash_output():
        content = await hash_result()
        return content

    @reactive.calc()
    async def hash_result() -> str:
        with ui.Progress() as p:
            p.set(message="Calculating...")

            value = input.value()
            return await asyncio.get_running_loop().run_in_executor(None, calc, value)


app = App(app_ui, server)



from shiny import reactive
from shiny.express import input, ui

ui.input_slider("controller", "Controller", min=0, max=1, value=0, step=1)
ui.input_checkbox("inCheckbox", "Input checkbox")


@reactive.effect
def _():
    # True if controller is odd, False if even.
    x_even = input.controller() % 2 == 1
    ui.update_checkbox("inCheckbox", value=x_even)



from shiny import App, Inputs, Outputs, Session, reactive, ui

app_ui = ui.page_fluid(
    ui.input_slider("controller", "Controller", min=0, max=1, value=0, step=1),
    ui.input_checkbox("inCheckbox", "Input checkbox"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.effect
    def _():
        # True if controller is odd, False if even.
        x_even = input.controller() % 2 == 1
        ui.update_checkbox("inCheckbox", value=x_even)


app = App(app_ui, server)



from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.playwright.expect._internal import _expect_nav_to_have_header_footer
from shiny.run import ShinyAppProc

navsets = [
    ("navset_pill", "navset_pill_a content", "navset_pill_b content"),
    ("navset_underline", "navset_underline_a content", "navset_underline_b content"),
    ("navset_tab", "navset_tab_a content", "navset_tab_b content"),
    ("navset_pill_list", "navset_pill_list_a content", "navset_pill_list_b content"),
    ("navset_card_pill", "navset_card_pill_a content", "navset_card_pill_b content"),
    ("navset_card_tab", "navset_card_tab_a content", "navset_card_tab_b content"),
    (
        "navset_card_underline",
        "navset_card_underline_a content",
        "navset_card_underline_b content",
    ),
]


def format_navset_name(navset_name: str) -> str:
    navset_name = navset_name.replace("_", " ")
    navset_name = navset_name.title()
    navset_name = navset_name.replace(" ", "")
    return navset_name


def test_navset_kitchensink(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    # Update the page size to be wider
    page.set_viewport_size({"width": 1500, "height": 800})

    page_navbar = controller.PageNavbar(page, "navsets_collection")

    # assert window title is same as page title if window title is not explicity set
    page_navbar.expect_window_title("Navsets kitchensink App")

    # for cases across all navsets
    for navset_name, default_content, selected_content in navsets:
        navset = controller.NavPanel(page, "navsets_collection", navset_name)
        navset.click()

        navset_default = getattr(controller, format_navset_name(navset_name))(
            page, f"{navset_name}_default"
        )
        navset_default._expect_content_text(default_content)
        navset_default.expect_value(f"{navset_name}_a")

        navset_selected = getattr(controller, format_navset_name(navset_name))(
            page, f"{navset_name}_selected"
        )
        navset_selected._expect_content_text(selected_content)
        navset_selected.expect_value(f"{navset_name}_b")

        navset_with_header_footer = getattr(
            controller, format_navset_name(navset_name)
        )(page, f"{navset_name}_with_header_footer")
        _expect_nav_to_have_header_footer(
            navset_with_header_footer.get_loc_active_content()
            .locator("..")
            .locator(".."),
            f"{navset_name}_header",
            f"{navset_name}_footer",
        )

        # assert header and footer contents
        expect(page.locator(f"#{navset_name}_header")).to_have_text(
            f"{navset_name}_with_header_footer header"
        )
        expect(page.locator(f"#{navset_name}_footer")).to_have_text(
            f"{navset_name}_with_header_footer footer"
        )
        if navset_name.startswith("navset_card"):
            navset_with_header_footer.expect_title(f"{navset_name}_with_header_footer")

            navset_card_underline_with_sidebar = getattr(
                controller, f"{navset_name.replace('_', ' ').title().replace(' ', '')}"
            )(page, f"{navset_name}_with_sidebar")
            navset_card_underline_with_sidebar.expect_sidebar(True)

        if navset_name in {"navset_card_underline", "navset_card_pill"}:
            navset_card_underline_placement = getattr(
                controller, f"{navset_name.replace('_', ' ').title().replace(' ', '')}"
            )(page, f"{navset_name}_placement")
            navset_card_underline_placement.expect_placement("below")

        if navset_name in {"navset_pill_list"}:
            navset_pill_list_default = controller.NavsetPillList(
                page, f"{navset_name}_default"
            )
            navset_pill_list_default.expect_well(True)
            navset_pill_list_default.expect_widths([4, 8])

            navset_pill_list_with_well = controller.NavsetPillList(
                page, f"{navset_name}_widths_no_well"
            )
            navset_pill_list_with_well.expect_well(False)
            navset_pill_list_with_well.expect_widths([10, 2])



from ._expect import (
    expect_not_to_have_attribute,
    expect_not_to_have_class,
    expect_to_have_class,
    expect_to_have_style,
    expect_not_to_have_style,
)
from ._expect_to_change import expect_to_change

__all__ = [
    "expect_to_change",
    "expect_not_to_have_attribute",
    "expect_to_have_class",
    "expect_not_to_have_class",
    "expect_to_have_style",
    "expect_not_to_have_style",
]



from __future__ import annotations

import re


def escape_to_var_name(x: str) -> str:
    """
    Given a string, escape it to a valid Python variable name which contains
    [a-zA-Z0-9_]. All other characters will be escaped to _<hex>_. Also, if the first
    character is a digit, it will be escaped to _<hex>_, because Python variable names
    can't begin with a digit.
    """
    encoded = ""
    is_first = True

    for char in x:
        if is_first and re.match("[0-9]", char):
            encoded += f"_{ord(char):x}_"
        elif re.match("[a-zA-Z0-9]", char):
            encoded += char
        else:
            encoded += f"_{ord(char):x}_"

        if is_first:
            is_first = False

    return encoded


def unescape_from_var_name(x: str) -> str:
    """
    Given a string that was escaped to a Python variable name, unescape it -- that is,
    convert it back to a regular string.
    """

    def replace_func(match: re.Match[str]) -> str:
        return chr(int(match.group(1), 16))

    decoded = re.sub("_([a-zA-Z0-9]+)_", replace_func, x)
    return decoded



from shiny.express import input, render, ui

ui.input_text("name", "Name", "Socrates")
ui.input_text("years", "Years", "470-399 BC")


@render.express
def person():
    with ui.card(class_="mt-3"):
        ui.h3(input.name())
        input.years()



import re

from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_output_image_kitchen(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    plotids_by_tab = {
        "mpl": [
            "mpl-plot_default",
            "mpl-plot_dom_size",
            "mpl-plot_decorator_size",
            "mpl-plot_native_size",
        ],
        "sns": [
            "sns-plot_default",
            "sns-plot_dom_size",
            "sns-plot_decorator_size",
            "sns-plot_native_size",
        ],
        "plotnine": [
            "plotnine-plot_default",
            "plotnine-plot_dom_size",
            "plotnine-plot_decorator_size",
            "plotnine-plot_native_size",
        ],
        "pil": [
            "pil-plot_default",
            "pil-plot_dom_size",
            "pil-plot_decorator_size",
        ],
    }

    for tab, plots in plotids_by_tab.items():
        page.click(f"a[data-toggle='tab'][data-value='{tab}']")
        for plotid in plots:
            img = controller.OutputPlot(page, plotid)
            # These assertions are mostly to ensure that the plots load before we
            # evaluate their sizes
            img.expect_inline(False)
            img.expect_img_src(re.compile(r"data:image/png;base64"), timeout=20000)

            rect = page.evaluate(
                f"() => document.querySelector('#{plotid} img').getBoundingClientRect()"
            )
            tolerance = 1e-4
            if plotid.startswith("sns"):
                # Not sure why but sns native-sized plot is a bit off
                tolerance += 2
            assert abs(rect["width"] - 300) <= tolerance
            assert abs(rect["height"] - 200) <= tolerance



from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_async_app(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    # Non-module version

    cb_show = controller.InputCheckbox(page, "show")
    expect(cb_show.loc).to_be_visible()
    expect(cb_show.loc).not_to_be_checked()

    loc = page.locator("#cond_message")
    expect(loc).to_be_hidden()

    cb_show.loc.check()
    expect(loc).to_be_visible()
    expect(loc).to_contain_text("Lorem ipsum dolor sit amet")

    # Module version

    cb_mod_show = controller.InputCheckbox(page, "mod-show")
    expect(cb_mod_show.loc).to_be_visible()
    expect(cb_mod_show.loc).not_to_be_checked()

    mod_loc = page.locator("#mod-cond_message")
    expect(mod_loc).to_be_hidden()

    cb_mod_show.loc.check()
    expect(mod_loc).to_be_visible()
    expect(mod_loc).to_contain_text("consectetur adipiscing elit")



from shiny.express import ui

ui.page_opts(title="Popover Kitchen sink", id="page_navbar")

ui.br()
ui.br()

with ui.popover(id="btn_popover_title", title="Popover title"):
    ui.input_action_button("btn", "A button")
    "Placement should be auto along with a title"

ui.br()
ui.br()
ui.br()
ui.br()

with ui.popover(id="btn_popover_top", placement="top"):
    ui.input_action_button("btn2", "A button")
    "Popover placement should be on the top"

ui.br()
ui.br()
ui.br()
ui.br()


with ui.layout_columns(col_widths=[4, 4]):

    with ui.popover(id="btn_popover_right", placement="right"):
        ui.input_action_button("btn3", "A button")
        "Popover placement should be on the right"

    with ui.popover(id="btn_popover_left", placement="left"):
        ui.input_action_button("btn4", "A button")
        "Popover placement should be on the left"



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def get_body_tag_name(card: controller.Card) -> str:
    body_tag_name = (
        card.loc_body.locator("*").nth(0).evaluate("el => el.tagName.toLowerCase()")
    )
    return body_tag_name


"""
For each card we want to test
Max Height and Min Height: The tests assert the max-height and min-height CSS properties applied to the card.
Header and Footer: The tests verify the presence and content of the header and footer elements within each card.
Body Element: The tests assert the tag name used for the body element within each card (e.g., <span>, <p>, <h1>, <h3>, <h5>).
Fullscreen Availability and State: The tests check whether the fullscreen feature is available for a particular card and verify its initial state (fullscreen or not). For cards with fullscreen support, the tests open and close the fullscreen mode and assert the expected behavior.
"""


def test_card_kitchensink(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    card = controller.Card(page, "card1")
    card.expect_max_height(None)
    card.expect_min_height(None)
    card.expect_height(None)
    card.expect_header("Check for header")
    card.expect_footer("Check for footer")
    card.expect_body(
        [
            "\nThis is the body of a card with default height w/ fullscreen",
        ]
    )
    card.expect_full_screen(False)
    card.set_full_screen(True)
    card.expect_full_screen(True)
    card.set_full_screen(False)
    card.expect_full_screen(False)

    card = controller.Card(page, "card2")
    card.expect_max_height(None)
    card.expect_min_height(None)
    card.expect_height(None)
    card.expect_header(None)
    card.expect_footer(None)
    card.expect_body(
        ["\nThis is the body without a header of a footer - No Fullscreen\n"]
    )
    assert get_body_tag_name(card) == "p"
    card.expect_full_screen(False)
    card.expect_full_screen_available(False)

    card = controller.Card(page, "card3")
    card.expect_max_height("500px")
    card.expect_min_height("200px")
    card.expect_header("Fill is False. Fullscreen is False")
    card.expect_footer(None)
    card.expect_body(["Max height and min height are set."])
    assert get_body_tag_name(card) == "h3"
    card.expect_full_screen_available(False)



# pyright: basic

from custom_component import custom_component

from shiny import App, render, ui

app_ui = ui.page_fluid(
    custom_component("myComponent"),
    ui.output_text("valueOut"),
)


def server(input, output, session):
    @render.text
    def valueOut():
        return f"Value from input is {input.myComponent()}"


app = App(app_ui, server)



from shiny.express import render, ui

# Set some Shiny page options
ui.page_opts(title="Hello Chat")


chat = ui.Chat(id="chat")
chat.ui()


@chat.on_user_submit
async def _():
    raise Exception("boom!")


"Message state:"


@render.code
def message_state():
    return str(chat.messages())



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_output_image_kitchen(page: Page, local_app: ShinyAppProc) -> None:

    page.goto(local_app.url)

    controller.OutputTextVerbatim(page, "all_txt").expect_value(
        "('a-3-flush', 'bx-3-first-flush', 'by-3-first-flush', 'bx-3-second-flush', "
        "'by-3-second-flush', 'c-3-flush', 'a-3-flushed', 'bx-3-first-flushed', "
        "'by-3-first-flushed', 'bx-3-second-flushed', 'by-3-second-flushed', "
        "'c-3-flushed')"
    )
    controller.OutputTextVerbatim(page, "flush_txt").expect_value(
        "('a-3-flush', 'bx-3-first-flush', 'by-3-first-flush', 'bx-3-second-flush', "
        "'by-3-second-flush', 'c-3-flush')"
    )
    controller.OutputTextVerbatim(page, "flushed_txt").expect_value(
        "('a-3-flushed', 'bx-3-first-flushed', 'by-3-first-flushed', "
        "'bx-3-second-flushed', 'by-3-second-flushed', 'c-3-flushed')"
    )
    # Session end messages have not flushed yet
    controller.OutputTextVerbatim(page, "session_end_txt").expect_value("[]")
    page.reload()
    # Session end messages have flushed
    controller.OutputTextVerbatim(page, "session_end_txt").expect_value(
        "['session ended - sync - test1', 'session ended - async - test2', 'session ended - async - test3', 'session ended - sync - test4']"
    )



from __future__ import annotations

from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_update_cell_value(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    pd_df = controller.OutputDataFrame(page, "pd_df")
    pd_df.expect_cell("PAL0708", row=0, col=0)

    pl_df = controller.OutputDataFrame(page, "pl_df")
    pl_df.expect_cell("PAL0708", row=0, col=0)

    pd_btn = controller.InputActionButton(page, "update_pd_btn")
    pl_btn = controller.InputActionButton(page, "update_pl_btn")

    # Update pandas
    pd_btn.click()
    pd_df.expect_cell("pandas - 1", row=0, col=0)
    pl_df.expect_cell("PAL0708", row=0, col=0)

    # Update polars
    pl_btn.click()
    pl_btn.click()
    pl_btn.click()

    pd_df.expect_cell("pandas - 1", row=0, col=0)
    pl_df.expect_cell("polars - 3", row=0, col=0)

    # Verify other cells do not change
    pd_df.expect_cell("PAL0708", row=1, col=0)
    pd_df.expect_cell("1", row=0, col=1)
    pd_df.expect_cell("2", row=1, col=1)

    pl_df.expect_cell("PAL0708", row=1, col=0)
    pl_df.expect_cell("1", row=0, col=1)
    pl_df.expect_cell("2", row=1, col=1)



import math
from functools import reduce
from operator import add
from statistics import mean

import numpy as np

WRIST = 0
THUMB_CMC = 1
THUMB_MCP = 2
THUMB_IP = 3
THUMB_TIP = 4
INDEX_FINGER_MCP = 5
INDEX_FINGER_PIP = 6
INDEX_FINGER_DIP = 7
INDEX_FINGER_TIP = 8
MIDDLE_FINGER_MCP = 9
MIDDLE_FINGER_PIP = 10
MIDDLE_FINGER_DIP = 11
MIDDLE_FINGER_TIP = 12
RING_FINGER_MCP = 13
RING_FINGER_PIP = 14
RING_FINGER_DIP = 15
RING_FINGER_TIP = 16
PINKY_MCP = 17
PINKY_PIP = 18
PINKY_DIP = 19
PINKY_TIP = 20


def hand_to_camera_eye(hands, detect_ok=False):
    # TODO: Pointing straight at the camera should not change (much) from the default
    # position
    # TODO: Use spherical coordinates instead of cartesian

    left_hand = hands["multiHandedness"][0]["index"] == 0
    hand = hands["multiHandLandmarks"][0]

    def rel_hand(start_pos: int, end_pos: int):
        return np.subtract(list(hand[start_pos].values()), list(hand[end_pos].values()))

    if detect_ok:
        # If the distance between the thumbtip and index finger tip are pretty close,
        # ignore the hand. (Using "OK" sign to pause hand tracking)
        ok_dist = np.linalg.norm(rel_hand(THUMB_TIP, INDEX_FINGER_TIP))
        ref_dist = np.linalg.norm(rel_hand(INDEX_FINGER_TIP, INDEX_FINGER_DIP))
        if ok_dist < ref_dist * 2:
            return None

    if not left_hand:
        normal_vec = np.cross(
            rel_hand(PINKY_MCP, WRIST),
            rel_hand(INDEX_FINGER_MCP, WRIST),
        )
    else:
        normal_vec = np.cross(
            rel_hand(INDEX_FINGER_MCP, WRIST),
            rel_hand(PINKY_MCP, WRIST),
        )
    normal_unit_vec = normal_vec / np.linalg.norm(normal_vec)

    def list_to_xyz(x):
        x = list(map(lambda y: round(y, 2), x))
        return dict(x=x[0], y=x[1], z=x[2])

    # Invert, for some reason
    normal_unit_vec = normal_unit_vec * -1.0
    normal_unit_vec[1] *= 2.0

    # Stay a consistent distance from the origin
    dist = 2
    magnitude = math.sqrt(reduce(add, [a**2 for a in normal_unit_vec]))
    normal_unit_vec = [a / magnitude * dist for a in normal_unit_vec]

    new_eye = list_to_xyz(normal_unit_vec)
    return {
        # Adjust locations to match camera position
        "x": new_eye["z"],
        "y": new_eye["x"],
        "z": new_eye["y"],
    }


def xyz_mean(points):
    return dict(
        x=mean([p["x"] for p in points]),
        y=mean([p["y"] for p in points]),
        z=mean([p["z"] for p in points]),
    )



from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.input_text("txt", "Enter the text to display below:", "delete me"),
    ui.row(
        ui.column(6, ui.code("ui.output_text()"), ui.output_text("text")),
        ui.column(
            6,
            ui.code("ui.output_text_verbatim(placeholder=True)"),
            ui.output_text_verbatim("verb", placeholder=True),
        ),
    ),
    ui.row(
        ui.column(6),
        ui.column(
            6,
            ui.code("ui.output_text_verbatim(placeholder=False)"),
            ui.output_text_verbatim("verb_no_placeholder", placeholder=False),
        ),
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.text
    def text():
        return input.txt()

    @render.text
    def verb():
        return input.txt()

    @render.text
    def verb_no_placeholder():
        return input.txt()


app = App(app_ui, server)



from shiny import App, Inputs, Outputs, Session, render, ui
from shiny.types import ImgData

app_ui = ui.page_fluid(ui.output_image("image"))


def server(input: Inputs, output: Outputs, session: Session):
    @render.image
    def image():
        from pathlib import Path

        dir = Path(__file__).resolve().parent
        img: ImgData = {"src": str(dir / "posit-logo.png"), "width": "100px"}
        return img


app = App(app_ui, server)



from shiny.express import ui

ui.page_opts(
    id="page_navbar_header_footer_fixed_top",
    header=ui.tags.div("Header", id="page_navbar_header"),
    footer=ui.tags.div("Footer", id="page_navbar_footer"),
    position="fixed-top",
    fluid=False,
)

with ui.nav_panel("Data"):
    "This page could be used to pick a dataset."

with ui.nav_panel("View"):
    "This page could be used to view the dataset."



from __future__ import annotations

import base64
import os
import sys
import typing

# `typing.Dict` sed for python 3.8 compatibility
# Can use `dict` in python >= 3.9
from typing import TYPE_CHECKING, Any, Callable, Literal, Optional, Union, cast

from htmltools import Tag, TagAttrValue, TagChild

from ._data_frame_utils._tbl_data import as_data_frame
from ._data_frame_utils._types import IntoDataFrame

if TYPE_CHECKING:

    from ..session._utils import RenderedDeps

from .. import _utils
from .. import ui as _ui
from .._docstring import add_example, no_example
from .._namespaces import ResolvedId
from .._typing_extensions import Self
from ..session import get_current_session, require_active_session
from ..session._session import DownloadHandler, DownloadInfo
from ..types import MISSING, MISSING_TYPE, ImgData
from ._try_render_plot import (
    PlotSizeInfo,
    try_render_matplotlib,
    try_render_pil,
    try_render_plotnine,
)
from .renderer import Jsonifiable, Renderer, ValueFn
from .renderer._utils import (
    imgdata_to_jsonifiable,
    rendered_deps_to_jsonifiable,
    set_kwargs_value,
)

__all__ = (
    "text",
    "code",
    "plot",
    "image",
    "table",
    "ui",
    "download",
)
# ======================================================================================
# RenderText
# ======================================================================================


@add_example(ex_dir="../api-examples/output_text")
@no_example("express")
class text(Renderer[str]):
    """
    Reactively render text.

    When used in Shiny Express applications, this defaults to displaying the text as
    normal text on the web page. When used in Shiny Core applications, this should be
    paired with :func:`~shiny.ui.output_text` in the UI.


    Parameters
    ----------
    inline
        (Express only). If ``True``, the result is displayed inline. (This argument is
        passed to :func:`~shiny.ui.output_text`.)

    Returns
    -------
    :
        A decorator for a function that returns a string.

    Tip
    ----
    The name of the decorated function (or ``@output(id=...)``) should match the ``id``
    of a :func:`~shiny.ui.output_text` container (see :func:`~shiny.ui.output_text` for
    example usage).

    See Also
    --------
    * :class:`~shiny.render.code`
    * :func:`~shiny.ui.output_text`
    """

    def auto_output_ui(
        self,
        *,
        inline: bool | MISSING_TYPE = MISSING,
    ) -> Tag:
        kwargs: dict[str, Any] = {}
        set_kwargs_value(kwargs, "inline", inline, self.inline)

        return _ui.output_text(self.output_id, **kwargs)

    def __init__(
        self,
        _fn: Optional[ValueFn[str]] = None,
        *,
        inline: bool = False,
    ) -> None:
        super().__init__(_fn)
        self.inline: bool = inline

    async def transform(self, value: str) -> Jsonifiable:
        return str(value)


# ======================================================================================
# RenderCode
# ======================================================================================


class code(Renderer[str]):
    """
    Reactively render text as code (monospaced).

    When used in Shiny Express applications, this defaults to displaying the text in a
    monospace font in a code block. When used in Shiny Core applications, this should be
    paired with :func:`~shiny.ui.output_code` in the UI.

    Parameters
    ----------
    placeholder
        (Express only) If the output is empty or ``None``, should an empty rectangle be
        displayed to serve as a placeholder? This does not affect behavior when the
        output is nonempty. (This argument is passed to :func:`~shiny.ui.output_code`.)


    Returns
    -------
    :
        A decorator for a function that returns a string.

    Tip
    ----
    The name of the decorated function (or ``@output(id=...)``) should match the ``id``
    of a :func:`~shiny.ui.output_code` container (see :func:`~shiny.ui.output_code` for
    example usage).

    See Also
    --------
    * :class:`~shiny.render.code`
    * :func:`~shiny.ui.output_code`
    """

    def auto_output_ui(
        self,
        *,
        placeholder: bool | MISSING_TYPE = MISSING,
    ) -> Tag:
        kwargs: dict[str, bool] = {}
        set_kwargs_value(kwargs, "placeholder", placeholder, self.placeholder)
        return _ui.output_code(self.output_id, **kwargs)

    def __init__(
        self,
        _fn: Optional[ValueFn[str]] = None,
        *,
        placeholder: bool = True,
    ) -> None:
        super().__init__(_fn)
        self.placeholder = placeholder

    async def transform(self, value: str) -> Jsonifiable:
        return str(value)


# ======================================================================================
# RenderPlot
# ======================================================================================


# It would be nice to specify the return type of ValueFn to be something like:
#   Union[matplotlib.figure.Figure, PIL.Image.Image]
# However, if we did that, we'd have to import those modules at load time, which adds
# a nontrivial amount of overhead. So for now, we're just using `object`.


@add_example(ex_dir="../api-examples/output_plot")
@no_example("express")
class plot(Renderer[object]):
    """
    Reactively render a plot object as an HTML image.

    Parameters
    ----------
    alt
        Alternative text for the image if it cannot be displayed or viewed (i.e., the
        user uses a screen reader).
    width
        Width of the plot in pixels. If ``None`` or ``MISSING``, the width will be
        determined by the size of the corresponding :func:`~shiny.ui.output_plot`. (You
        should not need to use this argument in most Shiny apps--set the desired width
        on :func:`~shiny.ui.output_plot` instead.)
    height
        Height of the plot in pixels. If ``None`` or ``MISSING``, the height will be
        determined by the size of the corresponding :func:`~shiny.ui.output_plot`. (You
        should not need to use this argument in most Shiny apps--set the desired height
        on :func:`~shiny.ui.output_plot` instead.)
    **kwargs
        Additional keyword arguments passed to the relevant method for saving the image
        (e.g., for matplotlib, arguments to ``savefig()``; for PIL and plotnine,
        arguments to ``save()``).

    Returns
    -------
    :
        A decorator for a function that returns any of the following:

        1. A :class:`matplotlib.figure.Figure` instance.
        2. An :class:`matplotlib.artist.Artist` instance.
        3. A list/tuple of Figure/Artist instances.
        4. An object with a 'figure' attribute pointing to a
           :class:`matplotlib.figure.Figure` instance.
        5. A :class:`PIL.Image.Image` instance.

    It's also possible to use the ``matplotlib.pyplot`` interface; in that case, your
    function should just call pyplot functions and not return anything. (Note that if
    the decorated function is async, then it's not safe to use pyplot. Shiny will detect
    this case and throw an error asking you to use matplotlib's object-oriented
    interface instead.)

    Tip
    ----
    The name of the decorated function (or ``@output(id=...)``) should match the ``id``
    of a :func:`~shiny.ui.output_plot` container (see :func:`~shiny.ui.output_plot` for
    example usage).

    See Also
    --------
    * :func:`~shiny.ui.output_plot`
    * :class:`~shiny.render.image`
    """

    def auto_output_ui(
        self,
        *,
        width: str | float | int | MISSING_TYPE = MISSING,
        height: str | float | int | MISSING_TYPE = MISSING,
        **kwargs: object,
    ) -> Tag:
        # Only set the arg if it is available. (Prevents duplicating default values)
        set_kwargs_value(kwargs, "width", width, self.width)
        set_kwargs_value(kwargs, "height", height, self.height)
        return _ui.output_plot(
            self.output_id,
            # (possibly) contains `width` and `height` keys!
            **kwargs,  # pyright: ignore[reportArgumentType]
        )
        # TODO: Deal with output width/height separately from render width/height?

    def __init__(
        self,
        _fn: Optional[ValueFn[object]] = None,
        *,
        alt: Optional[str] = None,
        width: float | None | MISSING_TYPE = MISSING,
        height: float | None | MISSING_TYPE = MISSING,
        **kwargs: object,
    ) -> None:
        super().__init__(_fn)
        self.alt = alt
        self.width = width
        self.height = height
        self.kwargs = kwargs

    async def render(self) -> dict[str, Jsonifiable] | Jsonifiable | None:
        is_userfn_async = self.fn.is_async()
        session = require_active_session(None)
        # Module support
        output_name = session.ns(self.output_id)
        width = self.width
        height = self.height
        alt = self.alt
        kwargs = self.kwargs

        inputs = session.root_scope().input

        # We don't have enough information at this point to decide what size the plot should
        # be. This is because the user's plotting code itself may express an opinion about
        # the plot size. We'll take the information we will need and stash it in
        # PlotSizeInfo, which then gets passed into the various plotting strategies.

        # Reactively read some information about the plot.
        pixelratio: float = typing.cast(
            float, inputs[ResolvedId(".clientdata_pixelratio")]()
        )

        # Do NOT call this unless you actually are going to respect the container dimension
        # you're asking for. It takes a reactive dependency. If the client hasn't reported
        # the requested dimension, you'll get a SilentException.
        def container_size(dimension: Literal["width", "height"]) -> float:
            result = inputs[
                ResolvedId(f".clientdata_output_{output_name}_{dimension}")
            ]()
            return typing.cast(float, result)

        non_missing_size = (
            cast(Union[float, None], width) if width is not MISSING else None,
            cast(Union[float, None], height) if height is not MISSING else None,
        )
        plot_size_info = PlotSizeInfo(
            container_size_px_fn=(
                lambda: container_size("width"),
                lambda: container_size("height"),
            ),
            user_specified_size_px=non_missing_size,
            pixelratio=pixelratio,
        )

        # Call the user function to get the plot object.
        x = await self.fn()

        # Note that x might be None; it could be a matplotlib.pyplot

        # Try each type of renderer in turn. The reason we do it this way is to avoid
        # importing modules that aren't already loaded. That could slow things down, or
        # worse, cause an error if the module isn't installed.
        #
        # Each try_render function should indicate whether it was able to make sense of
        # the x value (or, in the case of matplotlib, possibly it decided to use the
        # global pyplot figure) by returning a tuple that starts with True. The second
        # tuple element may be None in this case, which means the try_render function
        # explicitly wants the plot to be blanked.
        #
        # If a try_render function returns a tuple that starts with False, then the next
        # try_render function should be tried. If none succeed, an error is raised.
        ok: bool
        result: ImgData | None

        def cast_result(result: ImgData | None) -> dict[str, Jsonifiable] | None:
            if result is None:
                return None
            return imgdata_to_jsonifiable(result)

        if "plotnine" in sys.modules:
            ok, result = try_render_plotnine(
                x,
                plot_size_info=plot_size_info,
                alt=alt,
                **kwargs,
            )
            if ok:
                return cast_result(result)

        if "matplotlib" in sys.modules:
            ok, result = try_render_matplotlib(
                x,
                plot_size_info=plot_size_info,
                allow_global=not is_userfn_async,
                alt=alt,
                **kwargs,
            )
            if ok:
                return cast_result(result)

        if "PIL" in sys.modules:
            ok, result = try_render_pil(
                x,
                plot_size_info=plot_size_info,
                alt=alt,
                **kwargs,
            )
            if ok:
                return cast_result(result)

        # This check must happen last because
        # matplotlib might be able to plot even if x is `None`
        if x is None:
            return None

        raise Exception(
            f"@render.plot doesn't know to render objects of type '{str(type(x))}'. "
            + "Consider either requesting support for this type of plot object, and/or "
            + " explictly saving the object to a (png) file and using @render.image."
        )


# ======================================================================================
# RenderImage
# ======================================================================================
@add_example(ex_dir="../api-examples/output_image")
class image(Renderer[ImgData]):
    """
    Reactively render a image file as an HTML image.

    Parameters
    ----------
    delete_file
        If ``True``, the image file will be deleted after rendering.

    Returns
    -------
    :
        A decorator for a function that returns an :func:`~shiny.types.ImgData` object.

    Tip
    ----
    The name of the decorated function (or ``@output(id=...)``) should match the ``id``
    of a :func:`~shiny.ui.output_image` container (see :func:`~shiny.ui.output_image`
    for example usage).

    See Also
    --------
    * :func:`~shiny.ui.output_image`
    * :class:`~shiny.types.ImgData`
    * :class:`~shiny.render.plot`
    """

    def auto_output_ui(self, **kwargs: object):
        return _ui.output_image(
            self.output_id,
            **kwargs,  # pyright: ignore[reportArgumentType]
        )
        # TODO: Make width/height handling consistent with render_plot

    def __init__(
        self,
        _fn: Optional[ValueFn[ImgData]] = None,
        *,
        delete_file: bool = False,
    ) -> None:
        super().__init__(_fn)

        self.delete_file = delete_file

    async def transform(self, value: ImgData) -> dict[str, Jsonifiable] | None:
        src: str = value.get("src")
        try:
            with open(src, "rb") as f:
                data = base64.b64encode(f.read())
                data_str = data.decode("utf-8")
            content_type = _utils.guess_mime_type(src)
            value["src"] = f"data:{content_type};base64,{data_str}"
            return imgdata_to_jsonifiable(value)
        finally:
            if self.delete_file:
                os.remove(src)


# ======================================================================================
# RenderTable
# ======================================================================================


@add_example(ex_dir="../api-examples/output_table")
class table(Renderer[IntoDataFrame]):
    """
    Reactively render a pandas ``DataFrame`` object (or similar) as a basic HTML
    table.

    Consider using :class:`~shiny.render.data_frame` instead of this renderer, as
    it provides high performance virtual scrolling, built-in filtering and sorting,
    and a better default appearance. This renderer may still be helpful if you
    use pandas styling features that are not currently supported by
    :class:`~shiny.render.data_frame`.

    Parameters
    ----------
    index
        Whether to print index (row) labels. (Ignored for pandas :class:`~pandas.io.formats.style.Styler`
        objects; call ``style.hide(axis="index")`` from user code instead.)
    classes
        CSS classes (space separated) to apply to the resulting table. By default, we
        use `table shiny-table w-auto` which is designed to look reasonable with
        Bootstrap 5. (Ignored for pandas :class:`~pandas.io.formats.style.Styler` objects; call
        ``style.set_table_attributes('class="dataframe table shiny-table w-auto"')``
        from user code instead.)
    **kwargs
        Additional keyword arguments passed to ``pandas.DataFrame.to_html()`` or
        ``pandas.io.formats.style.Styler.to_html()``.

    Returns
    -------
    :
        A decorator for a function that returns any of the following:

        1. A pandas :class:`~pandas.DataFrame` object.
        2. A pandas :class:`~pandas.io.formats.style.Styler` object.
        3. Any object that has a `.to_pandas()` method (e.g., a Polars data frame or
           Arrow table).

    Tip
    ----
    The name of the decorated function (or ``@output(id=...)``) should match the ``id``
    of a :func:`~shiny.ui.output_table` container (see :func:`~shiny.ui.output_table`
    for example usage).

    See Also
    --------
    * :func:`~shiny.ui.output_table` for the corresponding UI component to this render function.
    """

    def auto_output_ui(self, **kwargs: TagAttrValue) -> Tag:
        return _ui.output_table(self.output_id, **kwargs)
        # TODO: Deal with kwargs

    def __init__(
        self,
        _fn: Optional[ValueFn[IntoDataFrame]] = None,
        *,
        index: bool = False,
        classes: str = "table shiny-table w-auto",
        border: int = 0,
        **kwargs: object,
    ) -> None:
        super().__init__(_fn)
        self.index: bool = index
        self.classes: str = classes
        self.border: int = border
        self.kwargs: dict[str, object] = kwargs

        # TODO: deal with kwargs collision with output_table

    async def transform(self, value: IntoDataFrame) -> dict[str, Jsonifiable]:
        import pandas
        import pandas.io.formats.style

        html: str
        if isinstance(value, pandas.io.formats.style.Styler):
            html = cast(  # pyright: ignore[reportUnnecessaryCast]
                str,
                value.to_html(**self.kwargs),  # pyright: ignore
            )
        else:
            if not isinstance(value, pandas.DataFrame):
                try:
                    nw_data = as_data_frame(value)
                except Exception as e:
                    raise TypeError(
                        "@render.table doesn't know how to render objects of type "
                        f"'{str(type(value))}'. Return eagar data frames that can "
                        "be handled by `narwhals`."
                    ) from e
                value = nw_data.to_pandas()

            html = cast(  # pyright: ignore[reportUnnecessaryCast]
                str,
                value.to_html(  # pyright: ignore
                    index=self.index,
                    classes=self.classes,
                    border=self.border,
                    **self.kwargs,  # pyright: ignore[reportArgumentType]
                ),
            )
        # Use typing to make sure the return shape matches
        ret: RenderedDeps = {"deps": [], "html": html}
        return rendered_deps_to_jsonifiable(ret)


# ======================================================================================
# RenderUI
# ======================================================================================
@add_example(ex_dir="../api-examples/output_ui")
class ui(Renderer[TagChild]):
    """
    Reactively render HTML content.

    Note: If you want to write your function with Shiny Express syntax, where the UI
    components are automatically captured as the code is evaluated, use
    :func:`~shiny.express.render.express` instead of this function.

    This function is used to render HTML content, but it requires that the funciton
    returns the content, using Shiny Core syntax.

    Returns
    -------
    :
        A decorator for a function that returns an object of type
        :class:`~htmltools.TagChild`.

    Tips
    ----
    The name of the decorated function (or ``@output(id=...)``) should match the ``id``
    of a :func:`~shiny.ui.output_ui` container (see :func:`~shiny.ui.output_ui` for
    example usage).

    See Also
    --------
    * :func:`~shiny.express.render.express`
    * :func:`~shiny.express.expressify`
    * :func:`~shiny.ui.output_ui`
    """

    def auto_output_ui(self) -> Tag:
        return _ui.output_ui(self.output_id)

    async def transform(self, value: TagChild) -> Jsonifiable:
        session = require_active_session(None)
        return rendered_deps_to_jsonifiable(
            session._process_ui(value),
        )


# ======================================================================================
# RenderDownload
# ======================================================================================
@add_example(ex_dir="../api-examples/download")
class download(Renderer[str]):
    """
    Decorator to register a function to handle a download.

    This decorator is used to register a function that will be called when the user
    clicks a download link or button. The decorated function may be sync or async, and
    should do one of the following:

    * Return a string. This will be assumed to be a filename; Shiny will return this
      file to the browser, and the downloaded file will have the same filename as the
      original, with an inferred mime type. This is the most convenient IF the file
      already exists on disk. But if the function must create a temporary file, then
      this method should not be used, because the temporary file will not be deleted by
      Shiny. Use the `yield` method instead.
    * `yield` one or more strings or bytestrings (`b"..."` or
      `io.BytesIO().getvalue()`). If strings are yielded, they'll be encoded in UTF-8.
      (This is better for temp files as after you're done yielding you can delete the
      temp file, or use a tempfile.TemporaryFile context manager) With this method, it's
      important that the `@render.download` decorator have a `filename` argument, as the
      decorated function won't help with that.

    Parameters
    ----------
    filename
        The filename of the download.
    media_type
        The media type of the download.
    encoding
        The encoding of the download.
    label
        (Express only) A label for the button. Defaults to "Download".

    Returns
    -------
    :
        The decorated function.

    See Also
    --------
    * :func:`~shiny.ui.download_button`
    * :func:`~shiny.ui.download_link`
    """

    def auto_output_ui(self) -> Tag:
        return _ui.download_button(
            self.output_id,
            label=self.label,
        )

    def __init__(
        self,
        fn: Optional[DownloadHandler] = None,
        *,
        filename: Optional[str | Callable[[], str]] = None,
        media_type: None | str | Callable[[], str] = None,
        encoding: str = "utf-8",
        label: TagChild = "Download",
    ) -> None:
        super().__init__()

        self.filename = filename
        self.media_type = media_type
        self.encoding = encoding
        self.label = label

        if fn is not None:
            self(fn)

    def __call__(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        fn: DownloadHandler,
    ) -> Self:
        # For downloads, the value function (which is passed to `__call__()`) is
        # different than for other renderers. For normal renderers, the user supplies
        # the value function. This function returns a value which is transformed,
        # serialized to JSON, and then sent to the browser.
        #
        # For downloads, the download button itself is actually an output. The value
        # that it renders is a URL; when the user clicks the button, the browser
        # initiates a download from that URL, and the server provides the file via
        # `session._downloads`.
        #
        # The `url()` function here is the value function for the download button. It
        # returns the URL for downloading the file.
        def url() -> str:
            from urllib.parse import quote

            session = require_active_session(None)
            # All download urls must be fully namespaced
            return f"session/{quote(session.id)}/download/{quote(session.ns(self.output_id))}?w="

        # Unlike most value functions, this one's name is `url`. But we want to get the
        # name from the user-supplied function.
        url.__name__ = fn.__name__

        # We invoke `super().__call__()` now, because it indirectly invokes
        # `Outputs.__call__()`, which sets `output_id` (and `self.__name__`), which is
        # then used below.
        super().__call__(url)

        # Register the download handler for the session. The reason we check for session
        # not being None or a stub session is because in Express, when the UI is
        # rendered, this function `render.download()()`  called once before any sessions
        # have been started.
        session = get_current_session()
        if session is not None and not session.is_stub_session():
            # All download objects are stored in the root session.
            # They must be fully namespaced
            session._downloads[session.ns(self.output_id)] = DownloadInfo(
                filename=self.filename,
                content_type=self.media_type,
                handler=fn,
                encoding=self.encoding,
            )

        return self

    async def transform(self, value: str) -> Jsonifiable:
        return value



from shiny.express import input, render, ui

with ui.accordion(id="acc"):
    for letter in "ABCDE":
        with ui.accordion_panel(f"Section {letter}"):
            f"Some narrative for section {letter}"


@render.code
def acc_val():
    return "input.acc(): " + str(input.acc())



from icons import arrow_up, piggy_bank

from shiny import App, ui

app_ui = ui.page_fluid(
    ui.layout_column_wrap(
        ui.value_box(
            "KPI Title",
            ui.h1(ui.HTML("$1 <i>Billion</i> Dollars")),
            ui.span(arrow_up, " 30% VS PREVIOUS 30 DAYS"),
            showcase=piggy_bank,
            theme="bg-gradient-orange-cyan",
            full_screen=True,
        ),
        ui.value_box(
            "KPI Title",
            ui.h1(ui.HTML("$1 <i>Billion</i> Dollars")),
            ui.span(arrow_up, " 30% VS PREVIOUS 30 DAYS"),
            showcase=piggy_bank,
            theme="text-green",
            showcase_layout="top right",
            full_screen=True,
        ),
        ui.value_box(
            "KPI Title",
            ui.h1(ui.HTML("$1 <i>Billion</i> Dollars")),
            ui.span(arrow_up, " 30% VS PREVIOUS 30 DAYS"),
            showcase=piggy_bank,
            theme="purple",
            showcase_layout="bottom",
            full_screen=True,
        ),
    )
)


app = App(app_ui, server=None)



from __future__ import annotations

import sys
from typing import Optional

from htmltools import Tag, TagAttrValue, TagFunction, TagList, wrap_displayhook_handler

from .. import ui as _ui
from .._docstring import add_example
from .._typing_extensions import Self
from ..session._utils import require_active_session
from ..types import MISSING, MISSING_TYPE, JsonifiableDict
from .renderer import AsyncValueFn, Renderer, ValueFn
from .renderer._utils import rendered_deps_to_jsonifiable, set_kwargs_value


@add_example(ex_dir="../api-examples/render_express")
class express(Renderer[None]):
    """
    Reactively render HTML content with output captured as in Shiny Express

    This is similar to :class:`~shiny.render.ui`, except that :class:`~shiny.render.ui`
    uses the return value from the the decorated function, whereas this function works
    like Shiny Express: as it executes each line of the decorated function, it calls
    :func:`~sys.displayhook()` on the result. This has the effect of "capturing" the
    output of each line.

    This decorator can be thought of as a combination of :class:`~shiny.render.ui` (for
    rendering and sending the dynamic UI to the client), and `~shiny.express.expressify`
    (for capturing the output of each line).

    Returns
    -------
    :
        A decorator for a function that returns `None`.

    See Also
    --------
    * ~shiny.render.ui
    * ~shiny.ui.output_ui
    * ~shiny.express.expressify
    * ~shiny.express.ui.hold
    """

    def auto_output_ui(
        self,
        *,
        inline: bool | MISSING_TYPE = MISSING,
        container: TagFunction | MISSING_TYPE = MISSING,
        fill: bool | MISSING_TYPE = MISSING,
        fillable: bool | MISSING_TYPE = MISSING,
        **kwargs: TagAttrValue,
    ) -> Tag:
        # Only set the arg if it is available. (Prevents duplicating default values)
        set_kwargs_value(kwargs, "inline", inline, self.inline)
        set_kwargs_value(kwargs, "container", container, self.container)
        set_kwargs_value(kwargs, "fill", fill, self.fill)
        set_kwargs_value(kwargs, "fillable", fillable, self.fillable)

        return _ui.output_ui(
            self.output_id,
            # (possibly) contains `inline`, `container`, `fill`, and `fillable` keys!
            **kwargs,  # pyright: ignore[reportArgumentType]
        )

    def __call__(self, fn: ValueFn[None]) -> Self:
        if fn is None:  # pyright: ignore[reportUnnecessaryComparison]
            raise TypeError("@render.express requires a function when called")

        async_fn = AsyncValueFn(fn)
        if async_fn.is_async():
            raise TypeError(
                "@render.express does not support async functions. Use @render.ui instead."
            )

        from ..express.expressify_decorator._expressify import expressify_unwrap_inplace

        fn = expressify_unwrap_inplace()(fn)

        # Call the superclass method with upgraded `fn` value
        super().__call__(fn)

        return self

    def __init__(
        self,
        _fn: Optional[ValueFn[None]] = None,
        *,
        inline: bool = False,
        container: Optional[TagFunction] = None,
        fill: bool = False,
        fillable: bool = False,
        **kwargs: TagAttrValue,
    ):
        super().__init__(_fn)
        self.inline: bool = inline
        self.container: Optional[TagFunction] = container
        self.fill: bool = fill
        self.fillable: bool = fillable
        self.kwargs: dict[str, TagAttrValue] = kwargs

    async def render(self) -> JsonifiableDict | None:
        results: list[object] = []
        orig_displayhook = sys.displayhook
        sys.displayhook = wrap_displayhook_handler(results.append)

        if self.fn.is_async():
            raise TypeError(
                "@render.express does not support async functions. Use @render.ui instead."
            )

        try:
            # Run synchronously
            sync_value_fn = self.fn.get_sync_fn()
            ret = sync_value_fn()
            if ret is not None:
                raise RuntimeError(
                    "@render.express functions should not return values. "
                    "Instead, @render.express dynamically renders every printable line "
                    "within the function body. (`None` is a valid return value.)"
                )
        finally:
            sys.displayhook = orig_displayhook
        if len(results) == 0:
            return None

        session = require_active_session(None)
        return rendered_deps_to_jsonifiable(
            session._process_ui(
                TagList(*results)  # pyright: ignore[reportArgumentType]
            )
        )



from __future__ import annotations

import modin.pandas as mpd  # pyright: ignore[reportMissingTypeStubs]
import narwhals.stable.v1 as nw
import palmerpenguins  # pyright: ignore[reportMissingTypeStubs]
import polars as pl
import pyarrow as pa

from shiny.express import render, ui

pd_df = palmerpenguins.load_penguins_raw().iloc[0:2, 0:2]

nw_df = nw.from_native(pd_df, eager_only=True)
pa_df = pa.table(pd_df)  # pyright: ignore[reportUnknownMemberType]
mpd_df = mpd.DataFrame(pd_df)
pl_df = pl.DataFrame(pd_df)


with ui.layout_columns():

    with ui.card():
        ui.h2("Original Pandas Data")

        @render.data_frame
        def pd_df_original():
            return render.DataGrid(
                data=pd_df,  # pyright: ignore[reportUnknownArgumentType]
                selection_mode="row",
            )

        "Selected row:"

        @render.data_frame
        def selected_pandas_row():
            return pd_df_original.data_view(selected=True)

        "Data type:"

        @render.code
        def pd_type():
            return str(type(pd_df))

        ui.markdown("`.data()` type:")

        @render.code
        def pd_data():
            return str(type(pd_df_original.data()))

        ui.markdown("`.data_view()` type:")

        @render.code
        def pd_data_view():
            return str(type(pd_df_original.data_view()))

        ui.markdown("`.data_view(selected=True)` type:")

        @render.code
        def pd_data_view_selected():
            return str(type(pd_df_original.data_view(selected=True)))

    with ui.card():
        ui.h2("Original narwhals Data")

        @render.data_frame
        def nw_df_original():
            return render.DataGrid(
                data=nw_df,
                selection_mode="row",
            )

        "Selected row:"

        @render.data_frame
        def selected_nw_row():
            return nw_df_original.data_view(selected=True)

        "Data type:"

        @render.code
        def nw_type():
            return str(type(nw_df))

        ui.markdown("`.data()` type:")

        @render.code
        def nw_data():
            return str(type(nw_df_original.data()))

        ui.markdown("`.data_view()` type:")

        @render.code
        def nw_data_view():
            return str(type(nw_df_original.data_view()))

        ui.markdown("`.data_view(selected=True)` type:")

        @render.code
        def nw_data_view_selected():
            return str(type(nw_df_original.data_view(selected=True)))

    with ui.card():
        ui.h2("Original PyArrow Data")

        @render.data_frame
        def pa_df_original():
            return render.DataGrid(
                data=pa_df,
                selection_mode="row",
            )

        "Selected row:"

        @render.data_frame
        def selected_pa_row():
            return pa_df_original.data_view(selected=True)

        "Data type:"

        @render.code
        def pa_type():
            return str(type(pa_df))

        ui.markdown("`.data()` type:")

        @render.code
        def pa_data():
            return str(type(pa_df_original.data()))

        ui.markdown("`.data_view()` type:")

        @render.code
        def pa_data_view():
            return str(type(pa_df_original.data_view()))

        ui.markdown("`.data_view(selected=True)` type:")

        @render.code
        def pa_data_view_selected():
            return str(type(pa_df_original.data_view(selected=True)))

    with ui.card():
        ui.h2("Modin Data")

        @render.data_frame
        def mpd_df_original():
            return render.DataGrid(
                data=mpd_df,
                selection_mode="row",
            )

        "Selected row:"

        @render.data_frame
        def selected_mpd_row():
            return mpd_df_original.data_view(selected=True)

        "Data type:"

        @render.code
        def mpd_type():
            return str(type(mpd_df))

        ui.markdown("`.data()` type:")

        @render.code
        def mpd_data():
            return str(type(mpd_df_original.data()))

        ui.markdown("`.data_view()` type:")

        @render.code
        def mpd_data_view():
            return str(type(mpd_df_original.data_view()))

        ui.markdown("`.data_view(selected=True)` type:")

        @render.code
        def mpd_data_view_selected():
            return str(type(mpd_df_original.data_view(selected=True)))

    with ui.card():
        ui.h2("Polars Data")

        @render.data_frame
        def pl_df_original():
            return render.DataGrid(
                data=pl_df,
                selection_mode="row",
            )

        "Selected row:"

        @render.data_frame
        def selected_pl_row():
            return pl_df_original.data_view(selected=True)

        "Data type:"

        @render.code
        def pl_type():
            return str(type(pl_df))

        ui.markdown("`.data()` type:")

        @render.code
        def pl_data():
            return str(type(pl_df_original.data()))

        ui.markdown("`.data_view()` type:")

        @render.code
        def pl_data_view():
            return str(type(pl_df_original.data_view()))

        ui.markdown("`.data_view(selected=True)` type:")

        @render.code
        def pl_data_view_selected():
            return str(type(pl_df_original.data_view(selected=True)))



import urllib.request
from pathlib import Path

import duckdb
import shinyswatch
from query import query_output_server, query_output_ui

from shiny import App, reactive, ui

folder = Path(__file__).parent
db_file = folder / "weather.db"


def load_csv(con, csv_name, table_name):
    csv_url = f"https://raw.githubusercontent.com/rfordatascience/tidytuesday/master/data/2022/2022-12-20/{csv_name}.csv"
    local_file_path = folder / f"{csv_name}.csv"
    urllib.request.urlretrieve(csv_url, local_file_path)
    con.sql(
        f"CREATE TABLE {table_name} AS SELECT * FROM read_csv_auto('{local_file_path}')"
    )


if not Path.exists(db_file):
    con = duckdb.connect(str(db_file), read_only=False)
    load_csv(con, "weather_forecasts", "weather")
    load_csv(con, "cities", "cities")
    con.close()

con = duckdb.connect(str(db_file), read_only=True)

button_style = {"style": "margin: 15px"}

app_ui = ui.page_fluid(
    ui.panel_title("DuckDB query explorer"),
    ui.row(
        ui.column(
            2,
            ui.row(
                button_style,
                ui.input_action_button("add_query", "Add Query"),
            ),
            ui.row(
                button_style,
                ui.input_action_button("show_meta", "Show Metadata"),
            ),
            ui.br(),
            ui.row(
                ui.markdown(
                    """
                    This app lets you explore a dataset using SQL and duckdb.
                    The data is stored in an on-disk [duckdb](https://duckdb.org/) database,
                    which leads to extremely fast queries.
                    """
                ),
            ),
        ),
        ui.column(
            10,
            ui.tags.div(
                query_output_ui("initial_query", remove_id="initial_query"),
                id="module_container",
            ),
        ),
    ),
    theme=shinyswatch.theme.flatly,
)


def server(input, output, session):
    mod_counter = reactive.value(0)

    query_output_server("initial_query", con=con, remove_id="initial_query")

    @reactive.effect
    @reactive.event(input.add_query)
    def _():
        counter = mod_counter.get() + 1
        mod_counter.set(counter)
        id = "query_" + str(counter)
        ui.insert_ui(
            selector="#module_container",
            where="afterBegin",
            ui=query_output_ui(id, remove_id=id),
        )
        query_output_server(id, con=con, remove_id=id)

    @reactive.effect
    @reactive.event(input.show_meta)
    def _():
        counter = mod_counter.get() + 1
        mod_counter.set(counter)
        id = "query_" + str(counter)
        ui.insert_ui(
            selector="#module_container",
            where="afterBegin",
            ui=query_output_ui(
                id, qry="SELECT * from information_schema.columns", remove_id=id
            ),
        )
        query_output_server(id, con=con, remove_id=id)

    @reactive.effect
    @reactive.event(input.rmv)
    def _():
        ui.remove_ui(selector="div:has(> #txt)")


app = App(app_ui, server)



from datetime import datetime

from shiny import reactive
from shiny.express import input, session, ui

ui.input_action_button("close", "Close the session")
ui.p(
    """If this example is running on the browser (i.e., via shinylive),
    closing the session will log a message to the JavaScript console
    (open the browser's developer tools to see it).
    """
)


def log():
    print("Session ended at: " + datetime.now().strftime("%H:%M:%S"))


_ = session.on_ended(log)


@reactive.effect
@reactive.event(input.close)
async def _():
    await session.close()



from __future__ import annotations

from colors import bg_color, fg_color
from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.playwright.expect._internal import expect_class_to_have_value
from shiny.run import ShinyAppProc


def test_colors_are_rgb() -> None:
    assert bg_color.startswith("rgb(")
    assert fg_color.startswith("rgb(")


def test_sidebar_bg_colors(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    for i in range(3, 5):
        content = page.locator(f"#m{i}")
        sidebar = page.locator(f"#s{i}")

        main_layout = sidebar.locator("..")

        open_desktop = "always" if i <= 2 else "open"
        open_mobile = "always" if i <= 2 else "closed"
        position_val = "left" if i % 2 == 1 else "right"

        expect(main_layout).to_have_attribute("data-open-desktop", open_desktop)
        expect(main_layout).to_have_attribute("data-open-mobile", open_mobile)

        expect_class_to_have_value(
            main_layout,
            "sidebar-right",
            has_class=position_val == "right",
        )

        expect(content).to_have_text(f"Main content - {i}")
        expect(sidebar).to_have_text(f"Sidebar content - {i}")

        # Only works if css file is loaded
        expect(sidebar).to_have_css("background-color", bg_color)
        expect(sidebar).to_have_css("color", fg_color)

    s3 = controller.Sidebar(page, "s3")
    s3.expect_position("left")
    s4 = controller.Sidebar(page, "s4")
    s4.expect_position("right")



from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.input_switch("somevalue", "Some value", False),
    ui.output_text("value"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.text
    def value():
        return input.somevalue()


app = App(app_ui, server)



# ------------------------------------------------------------------------------------
# A basic Shiny Chat example powered by Ollama.
# To run it, you'll need an Ollama server running locally.
# To download and run the server, see https://github.com/ollama/ollama
# To install the Ollama Python client, see https://github.com/ollama/ollama-python
# ------------------------------------------------------------------------------------

import ollama

from shiny.express import ui

# Set some Shiny page options
ui.page_opts(
    title="Hello Ollama Chat",
    fillable=True,
    fillable_mobile=True,
)

# Create and display empty chat
chat = ui.Chat(id="chat")
chat.ui()


# Define a callback to run when the user submits a message
@chat.on_user_submit
async def _():
    # Get messages currently in the chat
    messages = chat.messages(format="ollama")
    # Create a response message stream
    # Assumes you've run `ollama run llama3` to start the server
    response = ollama.chat(
        model="llama3",
        messages=messages,
        stream=True,
    )
    # Append the response stream into the chat
    await chat.append_message_stream(response)
